/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, 1994, 1995, 1996, 1997 Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME:         dbcodes.h
 *
 * SCCSINFO:            $Id: dbcodes.h,v 1.7 1997/04/12 20:33:21 lincks Exp $
 *
 * ORIGINAL AUTHOR(S):  ????
 *
 * MODIFICATIONS:
 *      1993-08-26 Martin Sjlin. Added definition of RCOOB
 *            for information sent after OOB. This information should
 *            be silently discarded in lowcom.c:LL_ReceiveStatus.
 *      1994-01-06 Martin Sjlin. Added definition of DISCONNECT_CODE
 *            and RECONNECT_CODE as preparation for disconnected WS.
 *      1994-06-16 Martin Sjlin. Added ACCESS_CODE for allowing
 *            change access for user id (0).
 *      < add modifications with name and date >
 *
 * DESCRIPTION:
 *  This file contains the communication codes exchanged between
 *  the dbms programs
 */

#ifndef DBCODES_H
#define DBCODES_H

/* should really be in liblincks.h, but since the dbs does */
/* history node (vs) splitting (into fragments) */
/* check SC_SYSTEM and VSNET in liblincks.h! */
#define SC_SYSTEM          "SYSTEM"
#define SC_CHAIN           "Chain"
#define SYSTEM_CHAIN       SC_SYSTEM,SC_CHAIN
#define VSNET              "VSNET"

/* Status codes from data base */
#define RCNYI     1   /* Unimplemented command. */
#define RCOK      2   /* Positive acknwoledge. */
#define RCLOGO    3   /* Close connection. */
#define RCBAD     4   /* Bad operation. */
#define RCUUP     5   /* Unable to unpack packet. */
#define RCERROR   6   /* Error # and error string follows. */
#define RCMOL     7   /* Here comes a molecule. */
#define RCBUFFER  8   /* Buffer molecule (not sent). */
#define RCMESSAGE 9   /* Here comes a pen message. */
#define RCOOB     10  /* Before this was a OOB */
#define RCNACCESS 11  /* No access for operation */
#define RCNONAME  12  /* No name found in server */

/* Command codes to data base */
#define ALIVE_CODE	1
#define LOGO_CODE	2
#define EDIT_CODE	3
#define STORE_CODE	4
#define RETRIEVE_CODE	5
#define RELEASE_CODE	6
#define QUERY_CODE	7
#define MESSAGE_CODE	8
#define DISCONNECT_CODE 9
#define RECONNECT_CODE  10
#define ACCESS_CODE     11
#define MAX_DB_CODE    (ACCESS_CODE)  /* Must be last <>_CODE! */

/*
 * The ALIVE_CODE is used for a workspace to "tickle" the database to
 * see that it is responding.
 * The LOGO_CODE is sent by the workspace at logout.
 * The EDIT_CODE is sent by the workspace to indicate that a new version
 * is created for an object.
 * The STORE_CODE is used when a node is stored.
 * The RETRIEVE_CODE is used when a workspace wants to retrieve a node.
 * The RELEASE_CODE is sent by the RELEASEOBJ function, to indicate to
 * the database that the workspace has released its memory previously
 * allocated for an object.
 * The QUERY_CODE is reserved for future query language processing at
 * the database side.
 * The MESSAGE_CODE is used when a workspace wants to get a pen message
 * (subsequent to an OOB signal).
 * The DISCONNECT_CODE is sent before disconnecting a WS, to enable storing
 * interface table on a file/DB.
 * The RECONNECT_CODE is sent when reconnecting a WS, to signal a read
 * in of a interface table from file/DB.
 * The ACCESS_CODE is sent when the client would like to change the
 * effective userame and userid. Return new uid if allow. 
 */

/* Socket definitions */
#define SPECBYTE 0xA5		/* Attention code from LINSRV */
#define TIME_OUT_LIMIT 36000	/* # of seconds before time out */


#define OOB_MESSAGE_CODE 0x1		/* pen message */
#define PEN_RESOLVE	 0		/* Resolve version node async. */
#define PEN_CHECK	 1		/* Node being edited by other user */
#define PEN_POLL	 2		/* User began editing in parallel */

/* Molecule Type Constants */
#define UNKNOWN_TYPE	0
#define OTHER		1
#define VSH		2
#define VS		3
#define VSE		4
#define VSS		5
/*
 * UNKNOWN_TYPE marks a node whose type is "unknown". This is used at
 * retrieve requests from sysfuns.
 * OTHER marks  that a node is a version node ("user's data").
 * VSH marks hidden history node (no PEN).
 * VS marks history node (PEN).
 * VSE marks history node at pre-edit.
 * VSS marks history node a final store.
 */

#endif	/* DBCODES_H */
