/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1994, 1995, 1996, 1997, 1997 Martin Sjlin.
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME: 	w3lincks.c
 *
 * SCCSINFO:		$Id: w3lincks.c,v 1.8 1997/04/12 20:33:28 lincks Exp $
 *
 * ORIGINAL AUTHOR:  Martin Sjlin, 1994-05-15
 *
 * MODIFICATIONS:
 *      <list mods with name and date>
 *
 * DESCRIPTION:
 *
 * 	This file contains the 'main' function for w3lincks is 
 * 	based on aimstart.c
 *
 *********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************
 * void main(unsigned int argc, char **argv)
 */

/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include "config.h"	/* includes system dependent includes */
#include "parser.h"
#define  NO_WIDGET
#include "aimtypes.h"
#include "machine.h"
#include "xconfig.h"

#include <errno.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netdb.h>			/* for tcp.c */

#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

/*********************************************************************
 * EXTERNALLY-AVAILABLE DATA FOUND IN THIS MODULE:
 *********************************************************************/
label sysroot;			/* A label for the system root object */
label userroot;			/* A label for the user object */
reference_structure *refstructs;/* List of all reference structures. */

/* for later purpose, e.g. see Hyper-G gateway ... */
int use_development_buttons = 0;/* whether to use all buttons on windows */
int nobuttons = 0;		/* whether to have no buttons at all */
int suppress = 0;		/* whether to suppress certain stuff */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
#include "f_aim_version.h"
#include "f_ll_version.h"
#include "f_ls_version.h"
#include "f_tcp.h"
#include "f_newexport.h"
#include "f_aimserver.h"
#include "f_aimbuildref.h"
#include "f_aimsubr.h"
#include "f_linckshandler.h"
#include "f_comhist.h"
#include "f_misc.h"

#include "f_aimpropagate.h"		/* propagatedisplay() */
#include "f_aimcmdstream.h"		/* do_commands() */

#include "f_pen_notify.h"

/********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern char LL_loginuser[];		/* dbcomm.c */
extern char *sys_errlist[];		/* errno.h */
extern int sys_nerr;			/* errno.h */
extern int errno;			/* errno.h */

/** aimbuildref.c **/
extern actionentry EXPANSION_ACTIONS[];	/* actions for GPD language */
extern char *EXPANSION_SYNTAX[];/* syntax for GPD language */

#ifdef COMHISTOFF
extern int global_comhistoff;
#endif

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
#define W3FILE "%s/w3.%s"
#define USAGE " -f -q -D -N -S -K keyfile -P [host:]port [-u username -p password] database"
#define COPYRIGHT "\
LINCKS, Copyright (C) 1994, 1995, 1996, 1997, 1997 Martin Sjlin.\n\
LINCKS comes with ABSOLUTELY NO WARRANTY.  It is free software, and you\n\
are welcome to redistribute it under the conditions of the GNU General\n\
Public License (see the file COPYING in the distribution or contact the\n\
Free Software Foundation)."

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static void w3shutdown P_((void));
static char *arginit P_(( char **vec ));
static RETSIGTYPE w3handler P_((int sig));

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/
static char *keyfile_format = W3FILE;
static int   foreground = 0;		/* if we do not fork off */
static int   quiet = 0;
static int   do_oob = 1;		/* for POLL_oob */
static int   suicide = FALSE;		/* when we get a kill signal ... */
static char *password = NULL;
static char *username = NULL;
static char *server_address = NULL;
static FILE *logfd = stderr;	        /* log file descriptor */

#ifdef COMHISTOFF
static char *ID = "@(#)w3lincks (command history disabled)";
#else
static char *ID = "@(#)w3lincks";
#endif

/*  */
/**********************************************************************
 * Function: void main(unsigned int argc, char **argv)
 *
 * The main function for xlincks
 *
 * Modifications:
 *      <list mods with name and date>
 */
int main(argc, argv)
  int argc;
  char **argv;
{
  char       *db_dir = (char *)NULL;
  int        nopoll = 0;		/* for usfuns.c */
  int        port;			/* port number we are listning on */
  char      *host = NULL;		/* remote host, optional */
  t_u32bits  addr;			/* optional addr (ip) for remote host*/
  long       key;			/* 'secret' key for connection */
  int        sockfd;			/* file descr for listning socket */
  int        penfd;			/* file descr for pen */
  fd_set     ready;			/* set of description to check on */
  int        maxfd;			/* no of max fd */
  char       logfile[MAXPATHLEN];

#ifdef COMHISTOFF
  global_comhistoff = 1;
#endif /* COMHISTOFF */

  db_dir = arginit(argv);

  if (password == NULL || username == NULL) {
     (void)fprintf(stderr, "Must have userid and password to login.\n");
     exit(1);
   }

  if (!quiet) {
    (void)printf("%s", COPYRIGHT);
    print_aim_version();
    print_liblincks_version();
    print_libshared_version();
  }

  (void)printf("Logging in to database %s.\n", db_dir);
  
  /* disconnect with environment *before* login into database */  
  if (!foreground)
    if(setup_daemon() < 0) {
      (void) fprintf(stderr,
		     "Failed to setup environment - %s\n",
		     sys_errlist[(errno > sys_nerr) ? 0 : errno]);
      exit(1);
    }

  if(NOPROMPT_LOGIN(username, password, db_dir) != SUCCESS) {
    (void) fprintf( stderr, 
		   "NOPROMPT_LOGIN: login failure to %s\n",
		   db_dir);
    exit(1);
  }

  /* reopen stderr for error logging */
  sprintf(logfile, "%s/w3.log", LOGDIR);
  if (!(logfd = freopen(logfile, "a", stderr))) {
    fprintf(stderr, 
	    "w3lincks: failed to open log file %s (%d)\n",
	    logfile, errno);
    exit(1);
  }
  
  /* zero out password & userid */
  while(*username != '\0') *username++=' '; /* strange, use zeros  ... */
  while(*password != '\0') *password++=' '; /* and watch the result */

 /* register the pen functions after login, as we can get pen
  * on following object when retrieving them a second time. */
  (void) RCB_REGISTERCALLBACK(RCB_PEN_RESOLVE, PEN_resolve, NULL);
  (void) RCB_REGISTERCALLBACK(RCB_PEN_CHECK, PEN_check, NULL);
  (void) RCB_REGISTERCALLBACK(RCB_PEN_POLL, PEN_poll, NULL);

  /* we do not want to have the function in usfuns.c doing any polling */
  (void) RCB_REGISTERCALLBACK(RCB_PEN_NOPOLL, NULL, (void *)&nopoll);

  /* Initialise the command history */
  if (CH_init("w3lincks")) {
    (void)fprintf(stderr, "Could not find the command history.\n");
    exit(1);
  }

  /* Update the command history with the LOGIN command */
  if (CH_update(CHSTR_LOGIN, 0, (char *)NULL, (char *)NULL, (char *)NULL,
		(char *)NULL, 0, (label *)NULL, (label *)NULL, (label *)NULL)) {
    (void)fprintf(stderr, "Could not access the command history.\n");
    exit(1);
  }

  if (GSR_GETSYSTEMROOT(&sysroot) != SUCCESS) {
    (void)fprintf(stderr, "%s: Couldn't load.\n", db_dir);
    exit(1);
  }

  if (GLI_GETLINKITEM(&sysroot, AIM_USERS, LL_loginuser, 1, &userroot)) {
    (void)fprintf(stderr, "%s:  Cannot retrieve user root.\n", db_dir);
    (void)fprintf(stderr, "Either user %s unknown or database corrupted.\n",
		  LL_loginuser);
    exit(1);
  }

  /* Initialise the parser module */
  init_parser();
  metagrammar();

  /* Define the GPD language actions and syntax */
  defineactions(EXPANSION_ACTIONS);
  definegrammar(EXPANSION_SYNTAX);

  /* Reset the list of all reference structures */
  refstructs = (reference_structure *)NULL;

  /* catch kill signal  */
  (void) Signal(SIGHUP,  SIG_IGN);	/* as a daemon ... */
  (void) Signal(SIGPIPE, SIG_IGN);	/* mm, ignore for now */
  (void) Signal(SIGTERM, w3handler);
  (void) Signal(SIGQUIT, w3handler);

  /* add server connection + add pen into select */
  if((sockfd = setup_socket(server_address, keyfile_format, 
			    &host, &port, &addr, &key)) < 0) {
    (void) fprintf(stderr,"w3lincks: failed to setup server socket\n");
    exit(1);
  }

  /* setup html server stuff */
  load_html();

  /* close down stdout */
  fclose(stdout);

  /* setup sockets and select for all stuff */
  maxfd = penfd = PEN_GET_HANDLE();
  maxfd = maxfd < sockfd ? sockfd : maxfd;
  maxfd++;

  FD_ZERO(&ready);			/* zero first */
  FD_SET(penfd, &ready);		/* on PEN */
  FD_SET(sockfd,&ready);		/* on acceppt */
  
  /* for ever do this ... */
  for(;;) {
    fd_set rmask;
    fd_set emask;			/* buggy, buggy .. */
    int newsock;
    struct timeval timev;
    int sc;

    if (suicide)  break;

    rmask = emask = ready;
    timev.tv_sec = 3600; timev.tv_usec = 0; 
    while((sc = select(maxfd, &rmask, (fd_set *) NULL, &emask, &timev)) < 0
	  && errno == EINTR) {
      if (suicide) w3shutdown();
      emask = rmask = ready;
      timev.tv_sec = 3600; timev.tv_usec = 0; 
    }

    if (suicide)  break;
#if defined(_AIX)			
    if (sc < 0 && errno == 0) /* strange ... 931103 msj */
      (void) fprintf(logfd, "FYI: Select(0) < 0 && errno = 0\n");
    else
#endif /* AIX */
    if ( sc < 0) 
      (void) fprintf(logfd, "Select() < 0 && errno = %d\n", errno);
    else if ( sc == 0 ) {
      /* timeout ... */
      if ((sc = SA_SENDALIVE()) != SUCCESS)
	(void) fprintf(logfd, "SA_SENDALIVE() = %d\n", sc);
    }
    else {
      if (FD_ISSET(penfd, &rmask))
	POLL_oob((void *) &do_oob);
      if (FD_ISSET(sockfd, &rmask)) 
	/* try to accept and create message */
	if((newsock = accept_call_if(sockfd, logfd, addr)) >= 0) 
	  create_html(newsock, logfd, key);
    }

    if (suicide)  break;

    /* mainloop part ... */
    for(propagatedisplay(); do_commands(); propagatedisplay())
      ;

  } /* for */

  w3shutdown();
  exit (0);			/* all's well.  this is for lint */
}

/*  */
/**********************************************************************
 * Function: static RETSIGTYPE w3handler(int sig)
 *
 * Modifications:
 *      <list mods with name and date>
 */
static RETSIGTYPE w3handler(sig)
  int sig;
{
  fprintf(logfd,"w3lincks: caught %d - closing down!\n", sig);
  suicide = TRUE;

}

/*  */
/**********************************************************************
 * Function: static void w3shutdown()
 *
 * Modifications:
 *      <list mods with name and date>
 */
static void w3shutdown()
{
  int i;

  if (delete_keyfile(keyfile_format) < 0)
    fprintf(logfd,
	    "w3lincks: failed to remove keyfile - %s\n",
	    sys_errlist[(errno > sys_nerr) ? 0 : errno]);

  for (i = 0; i < 6; i++) 
    switch (LOGOUT()) {
    case FAIL:
      (void)fprintf(logfd, "LOGOUT returned FAIL - Sleep then try again\n");
      sleep(3);
      break;

    case ERR_RCBAD:
      (void)fprintf(logfd, "LOGOUT returned ERR_RCBAD - Poll for PENs\n");
      (void) POLL_oob((void *) &do_oob);
      sleep(3);      
      break;

    case SUCCESS:
    default:
      exit(0);
    }
  exit(1);
}

/*  */
/**********************************************************************
 * Function: static char *arginit(char **vec)
 *
 * processes the command line options.
 *
 * Modifications:
 *      1993-12-25 Martin Sjlin. Added -u <user> -p <password>
 *                 as for t2lincks and the other tools.
 *      1994-04-27 Martin Sjlin. Added -
 *      <list mods with name and date>
 */
static char *arginit(vec)
  char **vec;
{
  char *db = (char *)NULL;
  char *argument;
  char *first = *vec;
  int user = 0, pword = 0;

  for (vec++; (argument = *vec) != (char *)NULL; vec++) {
    if (*argument == '-') {
      switch(*++argument) {
      case 'u' :
	if( (username = *++vec) == (char *) NULL) {
	  (void) fprintf( stderr,"Missing username after -u\n");
	  goto usage;
	}
	user = 1;
	break;
      case 'f' :
	foreground = 1;
	break;
      case 'p' :
	if ((password = *++vec) == (char *) NULL) {
	  (void) fprintf( stderr,"Missing password after -p\n"); 
	  goto usage;
	}
	pword = 1;
	break;
      case 'q':
	quiet = 1;
	break;
      case 'D':
        use_development_buttons = 1;
	break;
      case 'K':
        if ((keyfile_format = *++vec) == (char *) NULL) {
	  (void) fprintf( stderr, "Missing keyfile format after -K\n");
	  goto usage;
	} 
	break;

      case 'N':
	nobuttons = 1;
	break;
      case 'S':
	suppress = 1;
	break;
      case 'P':
        if ((server_address = *++vec) == (char *) NULL) {
	  (void) fprintf(stderr, 
			 "Missing server [host:]port number after -P\n");
	  goto usage;
	} 
	break;
      default:
	(void)fprintf(stderr, 
		      "Unknown xlincks switch: -%s.  Ignoring it.\n", argument);
      }
    }
    else {
      if (!db) 
        db = argument;
      else {
	(void)fprintf(stderr, "%s%s %s",
		"Multiple command line args which could\n",
		"be used as database directories.  Ignoring", argument);
      }
    }
  } 
  if (user != pword) {
    (void)fprintf(stderr, "You must use both -u and -p.\n");
    goto usage;
  }

  if (!server_address) {
    (void) fprintf(stderr,"No address to listening on specified\n");
    goto usage;
  }

  if (db || (db = (char *)getenv("LINCKSDBDIR")))
    return db;

 usage: 
  (void)fprintf(stderr, "Usage: %s %s\n", first, USAGE);
  exit(1);
}

/* strings used for popup_message */
#define INFO_MESSAGE "Information"
#define ERROR_MESSAGE "Error Message"
#define HELP_MESSAGE "Help Message"
#define PEN_MESSAGE "Parallel Editing Notification"

/*  */
/***********************************************************************
 * Function: void popup_message(va_alist)
 *
 * This function pops up a message in a popup window -- one of two
 * types:  either an error message (POPUP_MESS_ERROR), an
 * informational message (POPUP_MESS_INFO), or a help
 * message(POPUP_MESS_HELP).  Should be dismissed by the user.
 *
 * Modifications:
 *	<list mods here with name and date>
 */
#ifdef __STDC__
void popup_message(int message_type, char * fmt, ...)
#else
void popup_message(va_alist)
  va_dcl
#endif /* n __STDC__ */
{
  static char *mess_text[] = {
    ERROR_MESSAGE,
    INFO_MESSAGE,
    HELP_MESSAGE,
    PEN_MESSAGE
  };
  char message_str[1024];
  va_list ap;
#ifndef __STDC__
  int message_type;
  char *fmt;
  va_start(ap);
  message_type = va_arg(ap, int);
  fmt = va_arg(ap, char *);
#else
  va_start(ap, fmt);
#endif /* __STDC__ */
  (void)vsprintf(message_str, fmt, ap);
  va_end(ap);

  /* check on things... */
  if (message_str == (char *)NULL) {
    popup_message(POPUP_MESS_ERROR, "popup_message:  null message input");
    return;
  }

  if ((message_type != POPUP_MESS_HELP) &&
      (message_type != POPUP_MESS_ERROR) &&
      (message_type != POPUP_MESS_PEN) &&
      (message_type != POPUP_MESS_INFO)) {
    popup_message(POPUP_MESS_ERROR, "popup_message:  invalid message type");
    return;
  }

  (void) fprintf(logfd,"%s: %s", mess_text[message_type],message_str);
}

/*  */
/*******************************************************************
 * Function: void fatal_error(va_alist)
 *
 * This function prints out an error message and subsequently exits.
 * If a fatal error occurs, this is the function which should be used.
 *
 * Modifications:
 *	<list mods here with name and date>
 */
#ifdef __STDC__
void fatal_error(char *fmt, ...)
#else
void fatal_error(va_alist)
  va_dcl
#endif /* n __STDC__ */
{
  va_list ap;
#ifndef __STDC__
  char *fmt;
  va_start(ap);
  fmt = va_arg(ap, char *);
#else
  va_start(ap, fmt);
#endif /* __STDC__ */
  vfprintf(logfd, fmt, ap);
  va_end(ap);
  (void)fflush(logfd);
  exit(-1);
}

/*  */
/*******************************************************************
 * Function: void trace(va_alist)
 *
 * The following function acts like printf, except that it
 * don't necessarily have to write on stdout.
 *
 * Print trace or debug info
 *
 * Usage: function(char *fmt, ...)
 *
 * Modifications:
 *	<list mods here with name and date>
 */
#ifdef __STDC__
void trace(char *fmt, ...)
#else
void trace(va_alist)
  va_dcl
#endif
{
  va_list ap;
#ifndef __STDC__
  char *fmt;
  va_start(ap);
  fmt = va_arg(ap, char *);
#else
  va_start(ap, fmt);
#endif
  vfprintf(logfd,fmt, ap);
  va_end(ap);
}

/*  */
/*******************************************************************
 * Function: void short_error(va_alist)
 *
 * The following function acts like printf, except that it
 * don't necessarily have to write on stdout.
 *
 * Print an error message that is no longer than one line. For xlincks, these
 * are identical, but ttyaim also writes one-liners on the bottom line.
 *
 * Usage: function(char *fmt, ...)
 *
 * Modifications:
 *	<list mods here with name and date>
 */
#ifdef __STDC__
void short_error(char *fmt, ...)
#else
void short_error(va_alist)
  va_dcl
#endif
{
  va_list ap;
  static char message[1000];
#ifndef __STDC__
  char *fmt;
  va_start(ap);
  fmt = va_arg(ap, char *);
#else
  va_start(ap, fmt);  
#endif /* __STDC__ */
  vfprintf(logfd, fmt, ap);
  vsprintf(message, fmt, ap);
  va_end(ap);
  (void)fflush(logfd);
  popup_message(POPUP_MESS_ERROR, message);
}

/*  */
/*******************************************************************
 * Function: int pen_select_func (fd,pen_call_back,dopoll,idp)
 *
 * Add the pen_call_back to be called whenever we receive data
 * on the fd file descriptor.
 *
 * Modifications:
 *	<list mods here with name and date>
 */
int pen_select_func(fd, pen_call_back, dopoll, idp)
  int fd;
  void (* pen_call_back)();
  void *dopoll;
  XtInputId *idp;
{
#if 0
  /* make a list of functions, extra arguments and return a ptr to it */
  static int inited = FALSE;

  if (!inited) {
    FD_ZERO(&read_set);
    inited = TRUE;
  }

  FD_SET(fd,&read_set);
  p->func  = pen_call_back;
  p->extra = dopoll;

  /* if idp is non null set its value */
  if (idp != NULL)
    *idp = (void *) p;
#endif
  /* Currently we always returns true, but later ... */
  return( TRUE );
}

/*  */
/***********************************************************************
 * Function: void setplaceholder(infonode *np, int flag)
 *
 * Sets the value for a place-holder widget.
 *
 * Called by aimpropagate.c
 *
 * Modifications:
 *	<list mods here with name and date>
 */
void setplaceholder(np, flag)
  infonode *np;
  int flag;
{
  np->notindb = (flag)? 1 : 0;

  np->val.attsize = strlen(np->head->idname) + ((flag) ? 5 : 8);
  np->val.attvalue = (char *)malloc((ALLOC_T)np->val.attsize);
  if (np->val.attvalue != NULL) {
    if (flag)
      (void)sprintf(np->val.attvalue, "<<%s>>", np->head->idname);
    else
      (void)sprintf(np->val.attvalue, "empty: %s", np->head->idname);
  } 
  else 
    fatal_error("Couldn't allocate memory for attsize in setplaceholder\n");

}

/*  */
/*******************************************************************
 * Function: void uim_destroy(Widget widget)
 *
 * Destroy a widget.  Currently used to destroy widgets associated
 * with a reference structure being destroyed.
 *
 * Modifications:
 *	<list mods here with name and date>
 */
void uim_destroy(widget)
  Widget widget;
{
  if (widget == NULL)
    return;
#if 0
  XtDestroyWidget(widget);
#endif
}

/*  */
/*******************************************************************
 * Function: void uim_highlight(infonode *inp)
 *
 * mark widget if the widget has been "penned": the subject of
 * a parallel editing notification.
 *
 * Modifications:
 *	<list mods here with name and date>
 */
void uim_highlight(inp)
  infonode *inp;
{
  if (inp == NULL)
    return;
  if (is_penned((inp->index == LINKITEM) ? inp->lbl.vs : inp->obj.vs)) {
    (void)fprintf(logfd,"penned!\n");
  } else {
    (void)fprintf(logfd,"(de)penned!\n");
  }
}

/*  */
/*******************************************************************
 * Function: void mark_invisible_change(infonode *inp)
 *
 * mark widget if the widget has been changed ...
 *
 * Modifications:
 *	<list mods here with name and date>
 */
void mark_invisible_change(inp)
  infonode *inp;
{
  if ((inp == NULL) || (inp->widp == NULL))  return;
  (void) fprintf(logfd,"mark_invisible\n");
}

/*  */
/*******************************************************************
 * Function: void mark_changed(infonode *inp)
 *
 * mark widget if the widget has been changed ...
 *
 * Modifications:
 *	<list mods here with name and date>
 */
void mark_changed(inp)
  infonode *inp;
{
  if ((inp == NULL) || (inp->widp == NULL))  return;
  (void) fprintf(logfd,"mark_changed\n");
}

/*  */
/*******************************************************************
 * Function: void mark_inserted(infonode *inp)
 *
 * mark widget if the widget has been inserted ...
 *
 * Modifications:
 *	<list mods here with name and date>
 */
void mark_inserted(inp)
  infonode *inp;
{
  if ((inp == NULL) || (inp->widp == NULL))  return;
  (void) fprintf(logfd,"mark_inserted\n");

}

/*  */
/*******************************************************************
 * Function: void mark_deleted(infonode *inp)
 *
 * mark widget if the widget has been deleted
 *
 * Modifications:
 *	<list mods here with name and date>
 */
void mark_deleted(inp)
  infonode *inp;
{
  if ((inp == NULL) || (inp->widp == NULL))  return;
  (void) fprintf(logfd,"mark_deleted\n");
}

/*  */
/***********************************************************************
 * Function: char *makeshelltitle(identry *idp)
 *
 * Returns the name that should be used as the title for a window(!)
 *
 * Modifications:
 *     <list mods here with name and date>
 */
char *makeshelltitle(idp)
  identry *idp;
{
  static char buf[100];

  if (!(idp && idp->value && idp->value->val.attsize)) {
    /* Use GPD name for title */
    return idp->idname;
  }
  /* Take title from contents */
  if (idp->value->val.attsize < 30) {
    (void)strncpy(buf, idp->value->val.attvalue, idp->value->val.attsize);
    *(buf + idp->value->val.attsize) = '\0';
  } else {
    (void)strncpy(buf, idp->value->val.attvalue, 25);
    (void)strcpy((buf + 25), " ...");
  }
  return buf;

}

/* no need to rebuild any views yet ... */
void blddisplay(rsp)
  reference_structure *rsp;
{
  (void) fprintf(logfd,"blddisplay called!\n");
}


