/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1994, 1995, 1996, 1997 Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME: 	aimquery.c
 *
 * SCCSINFO:		$Id: aimquery.c,v 1.5 1997/04/12 20:33:55 lincks Exp $
 *
 * ORIGINAL AUTHOR:     Ralph Ronnquist
 *
 * MODIFICATIONS:
 *      <list mods with name and date>
 *
 * DESCRIPTION:
 *
 *********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************
 * 
 */

#include "config.h"     /* includes system dependent includes */
#include <parser.h>
#include <aimtypes.h>
#include <f_aimsubr.h>
#include <f_aimcmdstream.h>
#include "aimquery.h"

/*
   Process a selection statement relative an object, to decide whether
   that object belongs or not.
*/

char *QUERY_SYNTAX[] = {
    "<query> ::= [ [ <queryterm> while and ] while or ] => query",

    "<queryterm> ::= not <queryterm> => querynot !\
		<queryfactor> <relates> <queryfactor> => queryterm",

    "<queryfactor> ::= <queryattr> => queryfactor1 !\
		[ IMAGE ] =>queryfactor2 !\
		<queryid> => queryfactor3",

    "<queryattr> ::= '[ <queryid> : <queryid> ] => queryattr",

    "<queryid> ::= @ <queryattr> => queryindirect !\
	      <string> => querytext !\
	      <identifier> => querytext",

    "<relates> ::= \
		== => query_eq !\
		starts with => query_start !\
		contains => query_match",

/*
   The <queryaccess> language is the language to use as init string
   for a plural link item, excluding the surrounding double quotes.
   E.g. if the actual query statement is a root attribute X:Y, then
   the link item init string should look like
	"[X:Y]"
   while if the query is to compare target attribute X:Y with root
   attribute X:Y, the init stringmay look like:
	"\"[X:Y] == @[X:Y]\""
*/
    "<queryaccess> ::= <queryid> => noop",

    (char*)0
};

static query_cell *q_stack[500];
static int q_stkp;

#define push_q_stack(q)	q_stack[q_stkp++] = q
#define pop_q_stack	q_stack[--q_stkp]
#define q_stack_top	q_stack[q_stkp-1]

static query_cell *new_query_cell()
{
    query_cell *p = (query_cell*)malloc(sizeof(query_cell));
    p->next = 0;
    p->lhs = 0;
    p->rhs = 0;
    p->rel = QUERY_EQ;
    p->type = QUERY_ID;
    p->text = 0;
    return p;
}

static void make_expr(t)
    cell_type t;
{
    query_cell *q;

    if (q_stack_top->next) {
	q = pop_q_stack;
	push_q_stack(new_query_cell());
	q_stack_top->type = t;
	q_stack_top->lhs = q;
    }
}

static void linktop()
{
    /* Puts TOS after (TOS+1) */
    query_cell *x = pop_q_stack;
    q_stack_top->next = x;
}

static ACTION(query)
{
    int n_or = popint();
    int n_and;

    while (n_or-- > 1) {
	n_and = popint();
	while (n_and-- > 1) {
	    linktop();
	}
	make_expr(QUERY_AND);
	linktop();
    }
    make_expr(QUERY_OR);
    END_ACTION;
}

static ACTION(querynot)
{
    query_cell *lhs = pop_q_stack;
    push_q_stack(new_query_cell());
    q_stack_top->lhs = lhs;
    q_stack_top->type = QUERY_NOT;
    END_ACTION;
}

static ACTION(queryterm)
{
    query_cell *rhs = pop_q_stack;
    query_cell *lhs = pop_q_stack;
    char *rel = popdata();
    push_q_stack(new_query_cell());
    q_stack_top->lhs = lhs;
    q_stack_top->rhs = rhs;
    q_stack_top->type = QUERY_TERM;
    switch (rel[0]) {
      case 'E':
	q_stack_top->rel = QUERY_EQ;
	break;
      case 'M':
        q_stack_top->rel = QUERY_MATCH;
        break;
      case 'S':
	q_stack_top->rel = QUERY_START;
	break;
    }
    free(rel);
    END_ACTION;
}

static ACTION(queryfactor1)
{
    END_ACTION;
}

static ACTION(queryfactor2)
{
    push_q_stack(new_query_cell());
    q_stack_top->type = QUERY_IMAGE;
    END_ACTION;
}

static ACTION(queryfactor3)
{
    END_ACTION;
}

static ACTION(queryattr)
{
    query_cell *rhs = pop_q_stack;
    query_cell *lhs = pop_q_stack;
    push_q_stack(new_query_cell());
    q_stack_top->rhs = rhs;
    q_stack_top->lhs = lhs;
    q_stack_top->type = QUERY_ATTR;
    END_ACTION;
}

static ACTION(queryindirect)
{
    q_stack_top->type = QUERY_INDIRECT;
    END_ACTION;
}

static ACTION(querytext)
{
    push_q_stack(new_query_cell());
    q_stack_top->text = popdata();
    q_stack_top->type = QUERY_STRING;
    END_ACTION;
}

static ACTION(query_eq)
{
    pushdata(strdup("E"));
    END_ACTION;
}

static ACTION(query_match)
{
    pushdata(strdup("M"));
    END_ACTION;
}

static ACTION(query_start)
{
    pushdata(strdup("S"));
    END_ACTION;
}

static actionentry QUERY_ACTIONS[] =
  {
      {"query", query},
      {"querynot", querynot},
      {"queryterm", queryterm},
      {"queryfactor1", queryfactor1},
      {"queryfactor2", queryfactor2},
      {"queryfactor3", queryfactor3},
      {"queryattr", queryattr},
      {"queryindirect", queryindirect},
      {"querytext", querytext},
      {"query_eq", query_eq},
      {"query_match", query_match},
      {"query_start", query_start},
      {0,0}
  };

static int eval_relation(rel,lhs,rhs)
     cell_rel rel;
     char *lhs;
     char *rhs;
{
    switch (rel) {
      case QUERY_EQ:
	if ((!lhs) || (!rhs))
	  return 0;
	return (strcmp(lhs,rhs) == 0);
	
      case QUERY_MATCH:
	return (strstr(lhs,rhs) != 0);
      case QUERY_START:
	if (!rhs)
	  return 1;
	if (!lhs)
	  return 0;
	return (strncmp(lhs,rhs,strlen(rhs)) == 0);

      default:
	break;
    }
    return 0;
}

static char *eval_factor(obj,root,q)
     label *obj;
     label *root;
     query_cell *q;
{
    char *grp,*fld;
    attrval val;

    switch (q->type) {
      case QUERY_ID: /* id is constant string */
      case QUERY_STRING: /* factor is constant string */
	return strdup(q->text);
      case QUERY_INDIRECT: /* id is value of attribute */
	grp = eval_factor(obj,root,q->lhs);
	fld = eval_factor(obj,root,q->rhs);
	(void)GA_GETATTR(root,grp,fld,&val);
	free(grp);
	free(fld);
	return val.attvalue;
      case QUERY_ATTR: /* factor is value of attribute */
	grp = eval_factor(obj,root,q->lhs);
	fld = eval_factor(obj,root,q->rhs);
	(void)GA_GETATTR(obj,grp,fld,&val);
        free(grp);
        free(fld);
	return val.attvalue;
      case QUERY_IMAGE: /* factor is value of image */
	(void)GI_GETIMAGE(obj,&val);
	return val.attvalue;
      default:
	fprintf(stderr,"This should not be possible (1)\n");
	break;
    }
    return 0;
}

/*  */
/**********************************************************************
 * Function:
 *
 * Applies a query to an object.
 *
 * Modifications:
 *      <list mods with date and name>
 */
int evaluate_query(obj,root,q)
     label *obj;
     label *root;
     query_cell *q;
{
    char *lhs, *rhs;
    int x;

    switch (q->type) {
      case QUERY_NOT:
	if (evaluate_query(obj,root,q->lhs))
	  return 0;
        return 1;

      case QUERY_TERM:
	lhs = eval_factor(obj,root,q->lhs);
	rhs = eval_factor(obj,root,q->rhs);
	x = eval_relation(q->rel,lhs,rhs);
	free(lhs);
	free(rhs);
	return x;

      case QUERY_AND:
	for (q = q->lhs; q; q = q->next)
	  if (!evaluate_query(obj,root,q))
	    return 0;
	return 1;

      case QUERY_OR:
        for (q = q->lhs; q; q = q->next)
          if (evaluate_query(obj,root,q))
            return 1;
        return 0;
      default:
	fprintf(stderr,"This should not be possible (1)\n");
	break;
    }
    return 0;
}

/*  */
/**********************************************************************
 * Function: query_cell make_query(char *query_string)
 *
 * Translates the query string into a query structure
 *
 * Modifications:
 *	<list mods with date and name>
 */

static query_cell *make_query(query_string)
     char *query_string;
{
    static int inited = 0;

    if (!query_string)
      return 0;

    if (!inited) {
        inited = 1;
        defineactions(QUERY_ACTIONS);
        definegrammar(QUERY_SYNTAX);
    }

    if (DoParseExec(query_string,"query")) {
	fprintf(stderr,"Jibberish query: %s\n",query_string);
	return 0;
    }

    return pop_q_stack;
}

/*  */
/**********************************************************************
 * Function: query_cell establish_query(label *root,char *query_string)
 *
 * Translates the query string into a query structure
 *
 * Modifications:
 *	<list mods with date and name>
 */

query_cell *establish_query(root,query_string)
     label *root;
     char *query_string;
{
    static int inited = 0;

    if (!query_string)
      return 0;

    if (!inited) {
        inited = 1;
        defineactions(QUERY_ACTIONS);
        definegrammar(QUERY_SYNTAX);
    }

    if (DoParseExec(query_string,"queryaccess")) {
	fprintf(stderr,"Jibberish query access: %s\n",query_string);
	return 0;
    }

    return make_query(eval_factor(root,root,pop_q_stack));
}


/*  */
/**********************************************************************
 * Function: void free_query(query_cell *query)
 *
 * Translates the query string into a query structure
 *
 * Modifications:
 *	<list mods with date and name>
 */

void free_query(q)
     query_cell *q;
{
    query_cell *x;

    for (x=q; q; x=q) {
	q = q->next;
	if (x->lhs)
	  free_query(x->lhs);
	if (x->rhs)
	  free_query(x->rhs);
	if (x->text)
	  free(x->text);
	free(x);
    }
}

