/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1994, 1995, 1996, 1997, Martin Sjlin
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME:         aimserver.c
 *
 * SCCSINFO:	        $Id: aimserver.c,v 1.5 1997/04/12 20:33:56 lincks Exp $
 *
 * ORIGINAL AUTHOR(S):  Martin Sjlin 1994-05-14
 *
 * DESCRIPTION:
 *
 * This module contains functions to create X11 based server 
 * for using with newexport (initially).
 * 
 *
 * MODIFICATIONS:
 *	<list mods here with name and date>
 */

/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include "config.h"	/* includes system dependent includes */
#include "aimtypes.h"

#include "machine.h"			/* t_u16bits etc */
#include "xconfig.h"			/* DBDIR */
#include <sys/socket.h>			/* network stuff move later, TODO */
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <errno.h>
#include <netdb.h>			/* for struct hostent */
#include <sys/stat.h>			/* for chmod */

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
#include "f_aimserver.h"

/*********************************************************************
 * EXTERNALLY-AVAILABLE DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern char *sys_errlist[];	/* errno.h */
extern int sys_nerr;		/* errno.h */
extern int errno;		/* errno.h */
extern char LL_loginuser[];	/* set in dbcomm.c */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
#include "f_tcp.h"			/* for tcp_server */
#include "f_newexport.h"		/* for create_html etc */

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
typedef struct server_x{		/* server structure */
  struct server_x *next;		/* next, if any ... */
  int sockfd;				/* initial server socket */
  FILE *logfd;				/* file to print error and logging on */
  int port;				/* port number setup  */
  t_u32bits addr;			/* address limitations  */
  char *host;				/* host name limits */
  long key;				/* key, might not be used */
  XtInputId xid;			/* xid for this input source */
  void (*process)();			/* callback routine */
  void *extra;				/* application dependent stuff */
} server_callback_t;


/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static int setup_server P_((char *address,void (*func)(),char *keyfile));
static void server_cb   P_((void *p));
static long create_keyfile P_((int port, char *keyfile));

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/
static server_callback_t *servers = NULL;

/*  */
/*********************************************************************
 * Function: int newexport_server(char * port, char *keyfile)
 *
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
int html_server(address, keyfile)
  char *address;
  char *keyfile;
{
  static int done = FALSE;
  if (!done && address != NULL) {
    if ((done = setup_server(address, create_html, keyfile)))
      load_html();
    return(done);
  }
  return(FALSE);
}

/*  */
/*********************************************************************
 * Function: int accept_call_if(
 *
 * Accepts on the socket, check the remote address and shutdowns
 * the socket if it is not acceptable, and calls the real server
 * process.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
int accept_call_if(servfd, logfd, remote)
  int servfd;
  FILE *logfd;
  t_u32bits remote;
{
  int newsock;
  struct sockaddr_in client;
  int len = sizeof(client);
  struct hostent *hp = NULL;

  /* ok, we got it for reading lets do an accept */
  if((newsock = accept(servfd, (struct sockaddr *) &client, &len)) < 0) {
    (void) fprintf(logfd,
		   "accept_call_if: accept failed - %s\n",
		   sys_errlist[(errno > sys_nerr) ? 0 : errno]);
    return(-1);
  }

  /* check if we are restricting connction */
  if (remote !=  htonl(INADDR_ANY)) {
    if (remote != client.sin_addr.s_addr) {
      if ((hp = gethostbyaddr((char *) &client.sin_addr, 
			      sizeof(struct in_addr), 
			      AF_INET)) != NULL) 
	(void) fprintf(logfd,
		       "connection from non authorized server host - %s\n",
		       hp->h_name);
      else
	(void) fprintf(logfd, 
		       "connection from non server host - %s\n",
		       sys_errlist[(errno > sys_nerr) ? 0 : errno]);
      (void) shutdown(newsock,2);
      return(-1);
    }
  }

  return(newsock);
}

/*  */
/*********************************************************************
 * Function: static void server_cb(void *p)
 *
 * server_cb is installed as the callback function for accept event
 * on the initial setup server socket.  The only parameter `p' points
 * to a server_callback_t structure which contains all necessary
 * information such as: function to call, XtInputId id returned by
 * XtAddInput (via pen_select_func), limiting on accepting hosts
 * etc. 
 *
 * Accepts on the socket, check the remote address and shutdowns
 * the socket if it is not acceptable, and calls the real server
 * process.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static void server_cb(p)
  void *p;
{
  server_callback_t * cd = (server_callback_t *) p;
  int newsock;

  /* mm, let us just call the process function which should close etc the */
  if ((newsock = accept_call_if(cd->sockfd, cd->logfd, cd->addr)) >= 0)
    if (cd->process)
      (*cd->process)(newsock, cd->logfd, cd->key);
    else
      close(newsock);
  /* done for know */
}

/*  */
/*********************************************************************
 * Function: static int setup_server(char *address, void (*func()), char *keyfile)
 *
 * Sets up a server connection on `port', allowing only connections
 * from `host' (see server_cb), and having server_cb calling func
 * when we have a pending accept. 
 * 
 * If keyfile != NULL, create a key file containing key.
 *
 * If host != NULL, accept connection only from this host.
 *
 * Adds the `accepting' socket to the input mask using pen_select_fun,
 * load the Group:Field <-> shortname mapping, too.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static int setup_server(address, func, keyfile)
  char *address;
  void (*func)();
  char *keyfile;
{
  server_callback_t *cd;

  if ((cd=(server_callback_t *)malloc(sizeof(server_callback_t))) == NULL) {
    (void) fprintf(stderr,"server_setup: Failed to allocate memory\n");
    return(FALSE);
  }

  /* lookup host, create socket, create keyfile if necessary */
  cd->sockfd = 
    setup_socket(address, keyfile, &cd->host, &cd->port, &cd->addr, &cd->key);

  if (cd->sockfd < 0) {
    free((FREEPTR *) cd);
    return(FALSE);
  }

  /* now, here we add a call to select for X11 ... */
  if(pen_select_func (cd->sockfd, server_cb, cd, &cd->xid) < 0) {
    (void)fprintf(stderr, "server_setup: failed to add callback\n");
    close(cd->sockfd);
    free((FREEPTR *)cd);
    return(FALSE);
  }

  cd->process = func;
  cd->logfd   = stderr;

  /* and add it to our list of servers ... */
  cd->next = servers;
  servers = cd->next;

  return(TRUE);
}


/*  */
/*********************************************************************
 * Function: int setup_socket(char *address, char *keyfile, 
 *                   char **host, int *port, t_u32bits *addr, long *key);
 *
 * Sets up a server connection on `port', allowing only connections
 * 
 * If keyfile != NULL, create a key file containing key.
 *
 * If host != NULL, accept connection only from this host and set
 * * addr accordingly
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
int setup_socket(address, keyfile, host, port, addr, key)
  char *address;
  char *keyfile;

  char **host;
  int *port;
  t_u32bits *addr;
  long *key;
{
  int sockfd;
  char *p;

  /* check if we have full address - host:port number to bind to */

  *host = NULL;
  if ((p = strchr(address, ':')) != NULL) {
    if ((*host = strdup(address)) == NULL) { 
      (void)fprintf(stderr, "setup_socket: malloc failed\n");
      return(-1);
    }
    /* move into same place in copied string, chop it off */
    address = *host + (p - address);
    *address++ = '\0';
  }

  if ((*port = atoi(address)) <= 0) {
    (void) fprintf(stderr,"setup_socket: port no numbers\n");
    return(-1);
  }

  /* default */
  *addr = htonl(INADDR_ANY);

  /* look up host if given name */
  if (*host != NULL) 
    if (tcp_lookup(*host, addr, NULL) != SUCCESS) {
      (void) fprintf(stderr,  
		     "Failed to find a legal host addr for '%s'\n", *host);
      return(-1);
    }

  /* get a TCP socket for our use */
  if ((sockfd=tcp_server(tcp_error, *port, 5)) < 0) {
    (void) fprintf(stderr, "server_setup: Failed setup\n");
    return(-1);
  }

  /* open and create our server file ... */
  if (keyfile != NULL) 
    *key = create_keyfile(*port, keyfile);
  else
    *key = 0L;

  return(sockfd);
}


/*  */
/*********************************************************************
 * Function: static long create_keyfile(int port, char *keyfile)
 *
 * Create a key file containing key, this host, 
 * port name and username (LINCKS) on one line (format host:port:key:)
 * to enable cgi2lincks to connect to open the right connection.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static long create_keyfile(port, keyfile)
  int port;
  char *keyfile;
{
  long key = 0;
  FILE *fd;
  char ourhostname[256];
  char daemoninfo[MAXPATHLEN];	/* information file for cgi script */

  (void)sprintf(daemoninfo, keyfile, DBDIR, LL_loginuser);
#ifdef HAVE_RANDOM
  key = random() * getpid();
#else
  key = lrand48() * getpid();
#endif /* n HAVE_RANDOM */ 
  if ((fd = fopen(daemoninfo,"w")) == NULL) {
    (void) fprintf(stderr,
		   "create_keyfile: Failed to open %s for writing - %s\n", 
		   daemoninfo,
		   sys_errlist[(errno > sys_nerr) ? 0 : errno]);
    return(0L);      
  } 

  /* mm, get our own name */ 
  if(gethostname(ourhostname,sizeof(ourhostname)) < 0) {
    (void) fprintf(stderr,
		   "create_keyfile: gethostname failed - %s\n",
		   sys_errlist[(errno > sys_nerr) ? 0 : errno]);
    ourhostname[0]='\0';
  }
    
  fprintf(fd, "%s:%d:%ld:\n", ourhostname, (int) port, key);

  (void) fclose(fd);

  /* and change permissions, too */
  if (chmod(daemoninfo, S_IRUSR|S_IWUSR) < 0) {
    (void) fprintf(stderr,
		   "create_keyfile: failed to chmod on '%s' - %s\n", 
		   daemoninfo,
		   sys_errlist[(errno > sys_nerr) ? 0 : errno]);
  }
  return(key);
}

/*  */
/*********************************************************************
 * Function: int delete_keyfile(char *keyfile)
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
int delete_keyfile(keyfile)
  char *keyfile;
{
  char daemoninfo[MAXPATHLEN];	/* information file for cgi script */
  (void)sprintf(daemoninfo, keyfile, DBDIR, LL_loginuser);
  return(unlink(daemoninfo));
}

