/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1994, 1995, 1996, 1997, Martin Sjlin
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME:         newexport.c
 *
 * SCCSINFO:		$Id: newexport.c,v 1.6 1997/04/12 20:33:59 lincks Exp $
 *
 * ORIGINAL AUTHOR(S):  Martin Sjlin 1994-04-29
 *
 * DESCRIPTION:
 * This module contains function to create output using GPDs
 * 
 *
 * MODIFICATIONS:
 *	<list mods here with name and date>
 */

/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include "config.h"	/* includes system dependent includes */
#include "aimtypes.h"

#include <errno.h>

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
#include "f_newexport.h"

#ifdef XFORMAT
/*********************************************************************
 * EXTERNALLY-AVAILABLE DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern char *sys_errlist[];	/* errno.h */
extern int sys_nerr;		/* errno.h */
extern int errno;		/* errno.h */
extern label sysroot;		/* A label for the system root object */
extern label userroot;		/* A label for the user object */
extern reference_structure *refstructs;/* List of all reference structures. */
extern char LL_loginuser[];	/* set in dbcomm.c */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
#include "f_aimstruct.h"
#include "f_aimsubr.h"
#include "f_aimeditops.h"
#include "f_aimbuildref.h"
#include "f_aimstruct.h"
#include "f_usfuns.h"

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
#define HTML                 "HTML"

#define WWW_ROOT             "SYSTEM","WWWRoot"
#define WWW_NAME             "WWW","Username"
#define WWW_EMAIL            "WWW","E-mail address"         

#define EXPORT_AFTER         "ROOTAFTER"
#define EXPORT_BEFORE        "ROOTBEFORE"

#define DEFAULT_GPD          "node"
#define DEFAULT_FIELD_GPD    "folder"
#define USER_GPD             "user"
#define SEP_ITEM             ':'
#define SEP_GPD              '-'

#define LEVEL_1ST            -1
#define POS_1ST              -1

/* number of formats one can have associated with one widget */
#define NUM_FORMATS          100		
#define CHARSET              256	/* # of chars in characters set */

#define zerolen(str) ((str) == NULL ? zerostr : (str))

/* return values for the display_node_exec routine */
typedef enum { r_fail, r_integer, r_string, r_binary } t_ret;

/* return structre for display_node_exec and friends */
typedef struct x_formxret {
  t_ret what;				/* what kind of value */
  int allocated;			/* if allocated */
  int size;				/* size of value */
  union {
    int value;				/* integer value */
    char *str;				/* general pointer */
  } u;
} formxret;

/* for the remapping of GROUP:FIELD <-> shortname */
typedef struct { 
  char *name;
  char *group;
  char *field;
} t_shortpath;

/* structure definition for macro definition */
typedef struct {
   char *name;				/* macro name */
   formxentry *code;			/* macro code */
   } t_macro;

/* mode depend structure ... */
typedef struct x_mode {
  struct x_mode *next;			  /* next in list */
  char         *name;			  /* "name" of export type */
  char         *initstring;		  /* default init string */

  /* shortname <-> GROUP:FIELD */
  int           shortsize;		  /* entries shortname table */
  t_shortpath  *shortpath;		  /* shortname table */

  /* macro directory */
  int           macrosize;		  /* # of macros */
  t_macro      *macros;			  /* array of macro */

  /* character set remapping */
  char  remap_pbrk[CHARSET+1];		  /* contains characters to remap */
  char *remap[CHARSET];			  /* remap a character to a string */
  int   remap_state[CHARSET];		  /* number of states */
  int   remap_curr[CHARSET];		  /* current state */
  int   remap_maxsize;			  /* max size of any remap string */
} export_mode;

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static inline void reset_charset_state P_((char *mode));
static export_mode * find_export_mode P_((char *newmode));
static void          clear_printed   P_((infonode *np));
static char * remap_string P_(( char *str, char *newbuf, int *freeit));
static char * printf_part P_(( char *buf, char *group, char *field,
			       int pos ));
static char *make_path P_(( char *prefix, infonode *np, int len,
			   char **pbuf));
static void stack_formats   P_((infonode *np ));
static int firsttagname P_((void *extra, char *tag, int count, int length));
static char * find_default_gpd P_((label *gotit, char **gpdname_db));
static t_macro * find_macro P_((char *name));
static void find_shortname P_((infonode *np, char **group, char **field));
static int get_fullname P_((char *part, char **group, char **field,
			    int  *pos));
static int get_link P_((int doremap, char *part, 
			char **group, char **field, int  *pos, 
			char **gpd));
static int follow_path P_((char *path, label *gotit, char **prefix, 
			   char **gpdname, FILE *logfd));
static char x2c P_((char *what));
static int escape_url      P_((char *url, char *out));
static int unescape_url    P_((char *url, char *out));
static int ind P_((char *s, int c));
static int display_node_attr P_((int after, int level, int pos, 
				 infonode *np, int attr,
				 reference_structure *rsp));
static int free_it P_((formxret *ptr, int status));
inline static int display_node_export P_((int after, FILE *fd, formxentry *form,
					  int level, int pos, infonode *np,
					  int *attr, 
					  reference_structure *rsp,
					  formxret *rtn));
inline static int display_node_rel P_((int after, FILE *fd, formxentry *form,
				       int level, int pos, infonode *np,
				       int *attr, 
				       reference_structure *rsp,
				       formxret *rtn));
inline static int display_node_if P_((int after, FILE *fd, formxentry *form,
				      int level, int pos, infonode *np,
				      int *attr, 
				      reference_structure *rsp,
				      formxret *rtn));
inline static int display_node_case P_((int after, FILE *fd, formxentry *form,
					int level, int pos, infonode *np,
					int *attr, 
					reference_structure *rsp,
					formxret *rtn));
inline static int display_node_link P_((int after, FILE *fd, formxentry *form,
					int level, int pos, infonode *np,
					int *attr, 
					reference_structure *rsp,
					formxret *rtn));
inline static int display_node_format P_((int after, FILE *fd, formxentry *form,
					  int level, int pos, infonode *np,
					  int *attr, 
					  reference_structure *rsp,
					  formxret *rtn));
inline static int display_node_func P_((int after, FILE *fd, formxentry *form,
					int level, int pos, infonode *np,
					int *attr, 
					reference_structure *rsp,
					formxret *rtn));
static inline int display_node_var P_((int after, FILE *fd, formxentry *form,
				       int level, int pos, infonode *np,
				       int *attr, 
				       reference_structure *rsp,
				       formxret *rtn));
static int display_node_exec P_((int after, FILE *fd, formxentry *form,
				 int level, int pos, infonode *np,
				 int *attr, 
				 reference_structure *rsp,
				 formxret *rtn));
static void display_node    P_(( FILE *fd,
				reference_structure *rsp, 
				infonode *np,
				int level, int pos, int attr));
static int display_node_exec P_((int after, FILE *fd, formxentry *form, 
				 int level, int pos, infonode *np, 
				 int *attr, 
				 reference_structure *rsp, 
				 formxret *rtn));

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/
static formxentry *format_stack[NUM_FORMATS];	/* wild guess */
static int format_stack_ptr;

/* zero length string to fix pb with printf */
static char zerostr[] = "";	

/* mode setting */
static export_mode *mode_list = NULL;	  /* since we never will have many */
static export_mode *current_mode = NULL;  /* current mode */

/* stuff when parsing URL */
static label  our_userroot;		
static char  *context = NULL;		/* context for current document*/

/*  */
/**********************************************************************
 * Function: static inline void reset_charset_state(char *mode);
 * 
 * Try to find an export mode, if not found, create and initialize an export 
 * as well add to the existing one .
 * 
 * Modified:
 *	<list mods here with name and date>
 */
static inline void reset_charset_state(mode)
  char *mode;
{
  export_mode *p = find_export_mode(mode);
  int i;
  for(i=0; i < CHARSET; i++) 
    p->remap_curr[i] = 0;
}

/*  */
/**********************************************************************
 * Function: static export_mode * find_export_mode(char *name);
 * 
 * Try to find an export mode, if not found, create and initialize an export 
 * as well add to the existing one .
 * 
 * Modified:
 *	<list mods here with name and date>
 */
static export_mode * find_export_mode(newmode)
  char *newmode;
{
  export_mode *p;
  int i;

  for(p=mode_list; p != (export_mode *) NULL; p=p->next)
    if (strcmp(p->name, newmode) == 0)
      return(p);
    
  /* a new one! */
  if ((p = malloc((ALLOC_T) sizeof(export_mode))) == NULL)
      fatal_error("find_export_mode: malloc failed\n");

  p->name = strdup(newmode);
  p->initstring = NULL;
  p->shortsize = 0;
  p->shortpath = NULL;
  p->macrosize = 0;
  p->macros    = NULL;
  p->remap_maxsize = 0;  
  p->remap_pbrk[0]='\0';


  for(i=0; i < CHARSET; i++) {
    p->remap[i] = NULL;		
    p->remap_curr[i] = p->remap_state[i] = 0;
  }

  p->next = mode_list;
  mode_list = p;    

  return(p);
}

/*  */
/**********************************************************************
 * Function: void xblddisplay(reference_structure *rsp, FILE *fd, 
 *                            char *Context, char *mode)
 * 
 * Same as blddisplay, but only traverses the reference structures
 * and executes the action as specificed in ->xformats field.
 * 
 * Modified:
 *	<list mods here with name and date>
 */
void xblddisplay(rsp, fd, Context, mode)
  reference_structure *rsp;
  FILE *fd;
  char *Context;
  char *mode;
{
  if (rsp == (reference_structure *) NULL)  return;

  /* initialize pointer into stack of formats */
  format_stack_ptr = 0;	
  if (rsp->flags.bound_changed) {
    /* signal that it is bound */
    /* markbound(rsp, (int)rsp->flags.bound); */ 
    rsp->flags.bound_changed = 0;
  }

  /* if no earlier mode or current mode != input mode, load it */
  if (current_mode == NULL || strcmp(mode,current_mode->name) != 0) 
    current_mode = find_export_mode(mode);
  context = Context;
  reset_charset_state(mode);
  display_node(fd, rsp, rsp->target->value, LEVEL_1ST, POS_1ST, 0);
  clear_printed(rsp->target->value);
}

/*  */
/*********************************************************************
 * Function: static void clear_printed(infonode *np)
 *
 * Clear the printed attribute from all info nodes.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static void clear_printed(np)
  infonode *np;
{
  identry *idp;
  np->printed = 0;
  if ((idp = np->subnodes) != (identry *) NULL) 
    while (idp != (identry *) NULL) {
      for (np = idp->value; np != (infonode *) NULL; np=np->next) {
	clear_printed(np);
	np->printed = 0;
      }
      idp = idp->next;
    }
}

/*  */
/*********************************************************************
 * Function: void reset_charset(char *mode);
 *
 * Resets the remap arrays ...
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
void reset_charset(mode)
  char *mode;
{
  int i;
  export_mode *exp = find_export_mode(mode);

  /* lets go through and zero out everything */
  exp->remap_pbrk[0] = '\0';
  exp->remap_maxsize = 0;
  for(i=0; i < CHARSET; i++) {
    exp->remap[i] = NULL;
    exp->remap_state[i] = 0;
    exp->remap_curr[i] = 0;
  }
}

/*  */
/*********************************************************************
 * Function: void load_charset(char *mode, char *mapping)
 *
 * Loads the remap arrays from the string mapping. Earlier
 * mapping string MUST be free by the user (if allocated)
 *
 * format of mapping string:
 *
 *  [<state(byte)> <character(byte)> [<zero-terminated-string-state-0>]+]*
 *  '\0' 
 *
 * Notice that state byte must = 1, else it will be intepreted as ending
 * of remapping string ...
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
void load_charset(mode,mapping)
 char *mode;
 char *mapping;
{
  export_mode *exp;
  int i;
  int states;
  char * p = mapping;

  /* first, find the export mode */
  exp = find_export_mode(mode);
  
  /* now, traverse the mapping string */
  while(*p != '\0') {
    states=*p++;			  /* first state */
    exp->remap_state[(unsigned int)p[0]] = states;	  /* save number of states */
    exp->remap[(unsigned int)p[0]] = &p[1];		  /* ptr to first state */
    strncat(exp->remap_pbrk,&p[0],1);	  /* add first character to look for */
    p++;				  /* move to first string */

    /* states > 0 ... for each state move forward ... */
    for(i=0; states-- > 0; i=0) {
      while(*p++ != '\0')
	i++;
      /* keep track of longest string */
      if (i > exp->remap_maxsize) 
	exp->remap_maxsize = i;	
    }
  } /* end while */
}

/*  */
/*********************************************************************
 * Function: static char * remap_string(char *str, char *newbuf, int * freeit)
 *
 * Remaps the string `str', possible using `newbuf' for storage of remmaped
 * result if is != NULL. Frees `str' if `freeit' is TRUE and if we have
 * allocated new memory within this routinte.  Notice that any problems
 * allocating means that the routine returns `str'.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static char * remap_string(str, newbuf, freeit)
  char *str;
  char *newbuf;
  int  *freeit;
{
  char bigbuf[65536];
  register char *buf = bigbuf;
  register unsigned char *p = str;
  register int newsize = 0;
  char *last = &bigbuf[sizeof(bigbuf) - current_mode->remap_maxsize];

  /* if called with newbuf != NULL just use it and do not care */
  if (newbuf != NULL) 
    buf = newbuf;

  while(*p != '\0') {
    if (newbuf == NULL && buf > last)
	buf = bigbuf;			/* just wrap around  ... */

    switch (current_mode->remap_state[(unsigned int) p[0]]) {
    case 0: 
      *buf++ = *p++;
      break;

    case 1:  
      {
	register char *q = current_mode->remap[(unsigned int) p[0]];
	while((*buf++ = *q++)) newsize++;
	buf--;				/* skip zero */
	p++;				/* and take next */
	break;
      }

    default:  
      {
	unsigned int c = (unsigned int) p[0];
	register char *q = current_mode->remap[c];
	int i;

	/* skip as many state string as necessary */
	for( i=0; i < current_mode->remap_curr[c]; i++)
	  while(*q++ != '\0') ;

	if (++current_mode->remap_curr[c] == current_mode->remap_state[c])
	  current_mode->remap_curr[c]=0;

	while((*buf++ = *q++)) newsize++;	/* copy it and we are done */
	buf--;				/* skip zero */
	p++;				/* and move forward to next char */
	break;
      }
    } /* switch */
  } /* while */

  *buf = '\0';				/* zero end string */

  if (newsize == 0)			/* no changes ... */
    return(str);

  if (newbuf != NULL) {			/* always done here  */
    if (*freeit)
      free((FREEPTR *) str);
    *freeit = FALSE;
    return(newbuf);
  }

  newsize += strlen(str);		/* new length */

  /* if new length less than wrap around, we are done */
  if (newsize < sizeof(bigbuf) - current_mode->remap_maxsize) {	
    if (*freeit)
      free((FREEPTR *) str);
    *freeit = TRUE;
    return(strdup(bigbuf));
  }
    
  /* too long  - call ourself with the allocate size */
  if((newbuf = malloc((ALLOC_T) newsize)) == NULL)
    return(str);
  else {
    p = remap_string(str, newbuf, freeit);
    *freeit = TRUE;
    return(p);
  }
}

/*  */
/*********************************************************************
 * Function: static char * printf_part(char *buf, char * group, 
 *                           char *field, int pos)
 *
 * Constructs a partial path, e.g. [buf][:group][:pos] using the
 * input arguments (if they are present, e.g. != NULL), and
 * returns a pointer to construction. If the `buf'  == NULL
 * memory is allocated and the caller should free it, else
 * buf should point to storage enough.  Notice that pos is only
 * printed/included if > 0.
 *
 * Return: NULL upon failure to allocate memory or a pointer
 *         to constructed part.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static char * printf_part(buf, group, field, pos)
  char *buf;				/*  ptr to buffer */
  char *group;
  char *field;
  int pos;
{
  char *malloced = NULL;
  
  /*  malloc memory for space */
  if (buf == NULL) {
    size_t len = 16;				/* number space ... */
    len += group == NULL ? 1 : strlen(group);
    len += field == NULL ? 1 : strlen(field);
    if ((buf = malloced = malloc(len)) == NULL)
      return(NULL);
  } 

  if (group) 
    for(;*group; *buf++ = *group++);

  if (field) {
    *buf++ = SEP_ITEM;
    for(;*field; *buf++ = *field++);
  }

  /* if pos < 0 , do not print it .. */
  if (pos > 0)  {
    (void)sprintf(buf, "%c%d", SEP_ITEM, pos);
    buf += strlen(buf);			/* and move to end ... */
  }

  /* and a end-of-string */
  *buf = '\0';			

  /* and if we have malloced the buffer return start of it, */
  /* else the current position */
  return(malloced == NULL ? buf : malloced);
}

/*  */
/*********************************************************************
 * Function: static char *make_path(char *prefix, infonode *np, int len, 
 *                                 char **pbuf)
 *
 * Build a `access' path to reach the infonode np by first following
 * the parent links and at the same compute the length of the current
 * parts in the path. prefix is added in front of the path created,
 * and the first call to make_path should len containt the length of 
 * prefix.  *pbuf points to either a NULL, in the case the routine
 * allocated memory (callers should free), else *pbuf should point to 
 * storage enough to keep the whole path.
 *
 * Returns: NULL upon failure, or a pointer to the last written
 *          characters ... (e.g. ending zero)
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static char *make_path(prefix, np, len, pbuf)
  char *prefix;
  infonode *np;
  int len;
  char **pbuf;
{
  char *p = NULL;
  char *q = NULL;

  if (np == (infonode *) NULL) 
    return (NULL);

  /* when going upward, compute the length of the buffer */
  len += 1 + 1 + (np->group == NULL) ? 0 : (strlen(np->group));	
  len += 1 + 1 + (np->field == NULL) ? 0 : (strlen(np->field));
  len += 16;				/* np->pos at most ... */

  if (np->parent == (infonode *) NULL) {
    /* ok, we are at the top. */
    if (((p = *pbuf) != NULL) ||
        ((p = *pbuf = malloc(len)) == NULL))
      (void) fprintf(stderr, "make_path: Failed to allocate memory\n");
    else if (prefix != NULL)
      for(q=prefix; *q; *p++=*q++);
  } 
  else 
    p = make_path(prefix, np->parent, len, pbuf);
  
  /* if any failure we return NULL */
  if (p != NULL) {
    /* TODO .. this should be an index node ... add the output to */
    if (np->index == LINKITEM) {		
      char *group = np->group;
      char *field = np->field;

      /* do not write initial slash first time around ... */    
      if (p > *pbuf)
	*p++ = '/';				

      /* check for short name ... */
      find_shortname(np, &group, &field);

      p = printf_part(p, group, field, np->pos);

    }  /* is LINKITEM */
    else
      *p = '\0';
  }

  return(p);
}

/*  */
/*********************************************************************
 * Function: static int display_node_attr(int after, int level, int pos, 
 *                                        infonode *np, int attr, 
 *                                        reference_structure *rsp)
 *
 * the current given restriction as given in attr is compared to the
 * current state and if ANY of the restriction are true, we return
 * true,
 *
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static int display_node_attr(after, level, pos, np, attr, rsp)
 int after;
 int level;
 int pos;
 infonode *np;
 int attr;
 reference_structure *rsp;
{
  return(   ((attr & FX_ATTRAFTER)    && !after) 
	 || ((attr & FX_ATTRBEFORE)   &&  after) 
	 || ((attr & FX_ATTRFIRST)    && pos != 0) 
	 || ((attr & FX_ATTRNOTFIRST) && pos == 0)
	 || ((attr & FX_ATTRLAST)     && np->next != NULL) 
	 || ((attr & FX_ATTRNOTLAST)  && np->next == NULL)
	 || ((attr & FX_ATTRROOT)     && rsp->target->value != np) 
	 || ((attr & FX_ATTRLEAF)     && np->subnodes != NULL) 
	 || ((attr & FX_ATTRNODE)     && (np->subnodes == NULL 
					  || rsp->target->value == np))
	 || (!(attr & FX_ATTRPH)      && (np->notindb))
	 || ((attr & FX_ATTRONCE)     && (np->printed == 1))
	 );
}

/*  */
/*********************************************************************
 * Function: static int free_it(formxret *ptr, int status)
 *
 * return status and free any allocated memory.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static int free_it(ptr, status)
  formxret *ptr;
  int status;
{
  if (ptr->allocated) {
    if (ptr->u.str != NULL) {
      free((FREEPTR *) ptr->u.str);
      ptr->u.str = NULL;
    }
    ptr->allocated  = FALSE;
  }
  return(status);
}

/*  */
inline static int display_node_export(after, fd, form, level, pos, np, attr, rsp, rtn)
 int after;
 FILE *fd;
 formxentry *form;
 int level;
 int pos;
 infonode *np;
 int *attr;
 reference_structure *rsp;
 formxret *rtn;
{
  formxret format;
  t_macro * macro = NULL;

  if(display_node_attr(after, level, pos, np, FX_ATTRROOT, rsp))
    return(TRUE);

  if (!form->u.f.args)
    return(FALSE);

  if (!display_node_exec(after,fd,form->u.f.args,level,pos,np,attr,rsp,&format)
      || format.what != r_string)
    return(free_it(&format, FALSE));
    
  if (current_mode == NULL || strcmp(current_mode->name, format.u.str) != 0) {
    current_mode = find_export_mode(format.u.str);
#if 0
    /* TODO - check for object where, read default string */
    load_charset(current_mode->name,"\0");
    load_macro(current_mode->name,"\0"); 

    addformatspec(infonode *inp, label &obj, char *name); (aimbuildref.c)
      inp->xformats;
      notice that FORMATXS is group name    
      and MACRO name might the field -> one object/export type?
      and image => containts remap?
      and a links list of xformats obj which each maps
         a gpd name into a xformats spec ... 
         FORMATX (grp) : gpd name (field) => formats (attr)
         or link to another object with same grp/fld.
#endif
  }

  /* split it into before and after stuff */

  if (!display_node_attr(after, level, pos, np, FX_ATTRBEFORE, rsp)) {
    if (current_mode && current_mode->initstring != (char *) NULL)
      fprintf(fd, current_mode->initstring);
    macro = find_macro(EXPORT_BEFORE);
  } 
  else if (!display_node_attr(after, level, pos, np, FX_ATTRAFTER, rsp)) 
    macro = find_macro(EXPORT_AFTER);

  if (macro != NULL)
    return(free_it(&format, TRUE),
	   display_node_exec(after, fd, macro->code, level, pos, np, attr, rsp, rtn));
  else
    return(free_it(&format, TRUE));
}

/*  */
inline static int display_node_rel(after, fd, form, level, pos, np, attr, rsp, rtn)
 int after;
 FILE *fd;
 formxentry *form;
 int level;
 int pos;
 infonode *np;
 int *attr;
 reference_structure *rsp;
 formxret *rtn;
{
  formxret left, right;

  if (!form->u.f.args || !form->u.f.args->next) 
    return(FALSE);

  if (!display_node_exec(after,fd,form->u.f.args,level,pos,np,attr,rsp,&left))
    return(free_it(&left, FALSE));
  if (!display_node_exec(after,fd,form->u.f.args->next, 
			 level,pos,np,attr,rsp,&right)) 	
    return(free_it(&left, FALSE),free_it(&right, FALSE));      
    
  if(left.what == r_integer && right.what == r_integer) {
    rtn->allocated = FALSE;
    rtn->what = r_integer;
    switch(form->u.f.func) {
    case FX_OP_EQ : rtn->u.value = left.u.value == right.u.value;break;
    case FX_OP_NE : rtn->u.value = left.u.value != right.u.value;break;
    case FX_OP_LT : rtn->u.value = left.u.value <  right.u.value;break;
    case FX_OP_LE : rtn->u.value = left.u.value <= right.u.value;break;
    case FX_OP_GT : rtn->u.value = left.u.value >  right.u.value;break;
    case FX_OP_GE : rtn->u.value = left.u.value >= right.u.value;break;
    } /* switch */
    return(free_it(&left, TRUE),free_it(&right, TRUE));
  }  /* integers */
  else if (left.what == r_string && right.what == r_string) {
    rtn->allocated = FALSE;
    rtn->what = r_integer;
    switch(form->u.f.func) {
    case FX_OP_EQ : rtn->u.value = strcmp(left.u.str,right.u.str) == 0;break;
    case FX_OP_NE : rtn->u.value = strcmp(left.u.str,right.u.str) != 0;break;
    case FX_OP_LT : rtn->u.value = strcmp(left.u.str,right.u.str) <  0;break;
    case FX_OP_LE : rtn->u.value = strcmp(left.u.str,right.u.str) <= 0;break;
    case FX_OP_GT : rtn->u.value = strcmp(left.u.str,right.u.str) >  0;break;
    case FX_OP_GE : rtn->u.value = strcmp(left.u.str,right.u.str) >= 0;break;
    } /* switch */
    return(free_it(&left, TRUE),free_it(&right, TRUE));
  } /* strings */
  return(FALSE);
}
/*  */
inline static int display_node_if(after, fd, form, level, pos, np, attr, rsp, rtn)
 int after;
 FILE *fd;
 formxentry *form;
 int level;
 int pos;
 infonode *np;
 int *attr;
 reference_structure *rsp;
 formxret *rtn;
{
  formxret c;
  if (!form->u.f.args) 
    return(FALSE);
  if (!display_node_exec(after,fd,form->u.f.args,level,pos,np,attr,rsp,&c) 
      || (c.what != r_integer && c.what != r_string && c.what != r_binary))
    return(free_it(&c, FALSE));	
    
  if ((c.what == r_integer && c.u.value != 0) ||
      (c.what == r_binary && c.u.str != NULL) ||
      (c.what == r_string  && c.u.str != NULL)) { 
    /* true case */
    if (!form->u.f.args->next ||
	!display_node_exec(after, fd, form->u.f.args->next,
			   level, pos, np, attr, rsp, rtn))
      return(free_it(&c,FALSE),free_it(rtn,FALSE));
  }
  else {      
    /* false case */
    if (!form->u.f.args->next->next)
      return(free_it(&c,TRUE));		/* ok with on argument ... */
    if(!display_node_exec(after, fd, form->u.f.args->next->next, 
			  level, pos, np, attr, rsp, rtn))
      return(free_it(&c,FALSE),free_it(rtn,FALSE));
  }
  return(free_it(&c,TRUE));
}
/*  */
inline static int display_node_case(after, fd, form, level, pos, np, attr, rsp, rtn)
 int after;
 FILE *fd;
 formxentry *form;
 int level;
 int pos;
 infonode *np;
 int *attr;
 reference_structure *rsp;
 formxret *rtn;
{
  formxret i;
  formxentry *p;

  if (!form->u.f.args) 
    return(FALSE);

  if (!display_node_exec(after,fd,form->u.f.args,level,pos,np,attr,rsp,&i) 
      || i.what != r_integer)
    return(free_it(&i, FALSE));	/* only, after this it is an int */

  for(p=form->u.f.args->next; i.u.value > 0 && p != NULL; i.u.value--)
    p = p->next;
    
  if (i.u.value >= 0 && p!= NULL) 
    return(display_node_exec(after,fd,p,level,pos,np,attr,rsp,rtn));
  else
    return(TRUE);			/* just drop off the end if error */
}

/*  */
/* ATTR(GRP,FIELD) / ATTR(GRP,FIELD,POS,GRP,FIELD) => value */
/* LINK(GRP,FIELD,POS, ...) => IMAGE value ... */
inline static int display_node_link(after, fd, form, level, pos, np, attr, rsp, rtn)
 int after;
 FILE *fd;
 formxentry *form;
 int level;
 int pos;
 infonode *np;
 int *attr;
 reference_structure *rsp;
 formxret *rtn;
{    
  formxentry *p = form->u.f.args;
  label curr, nxt;
  attrval a;    

  /* at least one pair ... */
  if (!p || !p->next)
    return(FALSE);
    
  rtn->allocated = FALSE;
  rtn->what = r_string;
  rtn->u.str = NULL;
  rtn->size = 0;	  
    
  copylbl(&curr,&np->obj);
    
  while(p != NULL) {
    formxret group, field;
      
    /* if the last success ok .... */
    if (!display_node_exec(after,fd,p,level,pos,np,attr,rsp,&group) 
	|| group.what != r_string 
	|| p->next == NULL)
      return(free_it(&group,FALSE));

  back:
    if (!display_node_exec(after,fd,p->next,level,pos,np,attr,rsp,&field)
	|| group.what != r_string)
      return(free_it(&field,FALSE),free_it(&group,FALSE));
      
    /* we have group and a field ... */
    if (p->next->next == NULL) {
      /* attribute or image ? */
      if (form->u.f.func == FX_OP_ATTR) {
	attrval a;
	if (GA_GETATTR(&curr, group.u.str, field.u.str, &a) == SUCCESS) {
	  rtn->u.str = a.attvalue;
	  rtn->size = a.attsize;
	  rtn->allocated = TRUE;
	}
      }
      else if (form->u.f.func == FX_OP_LINK) {
	if (GLI_GETLINKITEM(&curr,group.u.str,field.u.str,1,&nxt)==SUCCESS) {
	  copylbl(&curr,&nxt);
	  (void) free_it(&field,TRUE); (void) free_it(&group,TRUE);
	  break; /* of while ... */
	}
      }
      return(free_it(&field,TRUE),free_it(&group,TRUE));
    } 
    else {
      /* we have "third" entry, this should an integer (pos)/string */
      formxret next;
	
      p = p->next->next;
      if(display_node_exec(after,fd,p,level,pos,np,attr,rsp,&next)) {
	if (next.what == r_string) {
	  /* it is a group entry */
	  if (GLI_GETLINKITEM(&curr,group.u.str,field.u.str, 1,&nxt) 
	      == SUCCESS) {
	    copylbl(&curr,&nxt);
	    (void) free_it(&group, FALSE); (void) free_it(&field, FALSE);
	    group.what      = next.what;
	    group.size      = next.size;
	    group.u.str     = next.u.str;
	    group.allocated = next.allocated;
	    goto back;
	  } else 
	    return(free_it(&next,TRUE),
		   free_it(&field,TRUE),
		   free_it(&group,TRUE));
	} /* r_string */
	/* it is a position */
	else if (next.what == r_integer) {
	  if (GLI_GETLINKITEM(&curr,group.u.str,field.u.str,
			      next.u.value,&nxt) == SUCCESS) {
	    copylbl(&curr,&nxt);
	    (void) free_it(&group, FALSE); (void) free_it(&field, FALSE); 
	    p=p->next;
	    continue;
	  } /* GLI */
	} /* r_integer */
      } /* dislay_exec */
      return(free_it(&next,FALSE),
	     free_it(&field,FALSE),
	     free_it(&group,FALSE));
    } /* end of chain */
  } /* while */
    
  /*  here, it should when a LINK item, and we want the value */
  if(form->u.f.func == FX_OP_ATTR) {
    (void)fprintf(stderr,
		  "FX_OP_ATTR: reached end of code - shouldn't\n");
    return(FALSE);
  }
      
  /* ok and now get image ... */
  if(GI_GETIMAGE(&curr,&a) == SUCCESS) {
    rtn->u.str = a.attvalue;
    rtn->size = a.attsize;
    rtn->what = r_binary;
    rtn->allocated = TRUE;
  }
  return(TRUE);
}

/*  */
inline static int display_node_format(after, fd, form, level, pos, np, attr, rsp, rtn)
 int after;
 FILE *fd;
 formxentry *form;
 int level;
 int pos;
 infonode *np;
 int *attr;
 reference_structure *rsp;
 formxret *rtn;
{
#define MAXARGC 8
  formxret argv[MAXARGC];		/* limits to 8 args today */
  int  argc = 0;
  int  attr_value = *attr;		/* used when checking for ONCE */
  formxentry *args = form->u.f.args;
  formxret frm;

  if (np->printed) return(TRUE); /* we have printed this ONCE */
  if (!form->u.f.args) 
    return(FALSE);

  /* get arguments - if first is attribute, check it ... */
  if (!display_node_exec(after,fd,args,level,pos,np,attr,rsp,&frm))
    return(free_it(&frm,FALSE));
      
  /* ok, it was an attribute or was converted into one ... */
  if (frm.what == r_integer) {
    if (frm.u.value == TRUE)	/* oops, done! */
      return(free_it(&frm, TRUE));
	
    /* do we have a second arg */
    if (!args->next || 
	!display_node_exec(after,fd,args->next,level,pos,np,attr,rsp,&frm))
      return(free_it(&frm, FALSE));
	
    /* ok, move part to next and make a copy of the attr value  */
    attr_value |= args->what == attribute ? args->u.attr : 0;
    args = args->next;
  }
      
  /* if the first argument is binary, we just write it out ... */
  if (frm.what == r_binary) {
    if (frm.size  > 0) {
      (void) fwrite(frm.u.str, frm.size, 1, fd);
      /* set printed flag *only* if we have ONCE enabled */
      if(attr_value & FX_ATTRONCE) np->printed = 1;
    }
    return(free_it(&frm,TRUE));
  }
      
  /* ok, let see. frm should be a format string ... */
  if (frm.what != r_string)
    return(free_it(&frm,FALSE));

  /* now, let start counting arguments, args->next is first ... */
  while (argc < MAXARGC && args->next != NULL) {

    /* if it fails, free all computed results ...*/
    if (!display_node_exec(after,fd,args->next,level,pos,np,attr,rsp,&argv[argc])
	|| argv[argc].what != r_string) {
      while(argc >= 0) 
	(void) free_it(&argv[argc--], FALSE);
      return(FALSE);
    }

    /* fix for (null) print outs */
    if (argv[argc].u.str == NULL) 
      argv[argc].u.str = (char *) zerostr;
    else if (!(attr_value & FX_ATTRNOREMAP))
      argv[argc].u.str = remap_string(argv[argc].u.str, 
				      NULL, 
				      &(argv[argc].allocated));
    argc++;
    args = args->next;
  }

  /* if no arguments use the infonode value  */
  if (argc == 0) {
    int freeit = FALSE;
    char *p;

    p = attr_value & FX_ATTRNOREMAP ? zerolen(np->val.attvalue) :
      remap_string(zerolen(np->val.attvalue), NULL, &freeit);

    fprintf(fd, frm.u.str, p);

    if (freeit) 
      free((FREEPTR *) p);
  }
  else {
    int i;

    /* eaiser and safer to zero out all pointers and then call */
    for(i=argc; i < MAXARGC; i++)
      argv[i].u.str = zerostr;

    fprintf(fd, frm.u.str, argv[0].u.str, argv[1].u.str,
	    argv[2].u.str, argv[3].u.str, argv[4].u.str,
	    argv[5].u.str, argv[6].u.str, argv[7].u.str);

  } /* argc != 0 */
	
  /* set printed flag *only* if we have ONCE enabled */
  if(attr_value & FX_ATTRONCE) np->printed = 1;
	
  /* and last free the arguments .. */
  while( argc-- > 0) 
    (void)free_it(&argv[argc], FALSE);

  return(free_it(&frm, TRUE));
}

/*  */
inline static int display_node_func(after, fd, form, level, pos, np, attr, rsp, rtn)
 int after;
 FILE *fd;
 formxentry *form;
 int level;
 int pos;
 infonode *np;
 int *attr;
 reference_structure *rsp;
 formxret *rtn;
{
  switch(form->u.f.func) {
  case FX_OP_MACRO: {
    t_macro *macro;
    if(!form->u.f.args ||
       !display_node_exec(after,fd,form->u.f.args,level,pos,np,attr,rsp,rtn) ||
       rtn->what != r_string)
      return(free_it(rtn, FALSE));
    macro = find_macro(rtn->u.str);
    if (macro == NULL)
      return(free_it(rtn, FALSE));
    else {
      return(free_it(rtn, FALSE),
	     display_node_exec(after, fd, macro->code, level, pos, np, attr, rsp, rtn));
    }
  }

  case FX_OP_SET_ATTR :
  case FX_OP_ADD_ATTR :
  case FX_OP_DEL_ATTR :
    break;
    
  case FX_OP_EXPORT:
    return(display_node_export(after, fd, form, level, pos, np, attr, rsp, rtn));

  case FX_OP_FORMAT : 
    return(display_node_format(after, fd, form, level, pos, np, attr, rsp, rtn));
    
  case FX_OP_IF : 
    return(display_node_if(after, fd, form, level, pos, np, attr, rsp, rtn));
    
  case FX_OP_CASE : 
    return(display_node_case(after, fd, form, level, pos, np, attr, rsp, rtn));

  case FX_OP_LINK : 
  case FX_OP_ATTR : 
    return(display_node_link(after, fd, form, level, pos, np, attr, rsp, rtn));
    
  case FX_OP_EQ :
  case FX_OP_NE :
  case FX_OP_LT :
  case FX_OP_LE :
  case FX_OP_GT :
  case FX_OP_GE : 
    return(display_node_rel(after, fd, form, level, pos, np, attr, rsp, rtn));
    
  case FX_OP_I2STR : /* if binary and ended by a zero ... */
    if(!form->u.f.args) 
      return(FALSE);
    if(!display_node_exec(after,fd,form->u.f.args,level,pos,np,attr,rsp,rtn)) 
      return(free_it(rtn, FALSE));
    if (rtn->what == r_binary && 
	rtn->size > 0 &&
	rtn->u.str[rtn->size-1] == '\0')
      rtn->what = r_string;
    break;

  case FX_OP_N2STR : /* if an integer, convert to string ... */
    if(!form->u.f.args) 
      return(FALSE);
    if(!display_node_exec(after,fd,form->u.f.args,level,pos,np,attr,rsp,rtn) ||
       rtn->what != r_integer)
      return(free_it(rtn, FALSE));
    {
      char buf[64];
      sprintf(buf,"%ld",(long) rtn->u.value);
      rtn->what      = r_string;
      rtn->u.str     = strdup(buf);
      rtn->size      = strlen(rtn->u.str);
      rtn->allocated = TRUE;
    }
    break;
    
  case FX_OP_ESCAPE : /* escape for a URL ... */
    if(!form->u.f.args) 
      return(FALSE);
    if(!display_node_exec(after,fd,form->u.f.args,level,pos,np,attr,rsp,rtn) ||
       rtn->what != r_string ||
       rtn->u.str == NULL)
      return(free_it(rtn, FALSE));
    {
      char  buf[16384];
      char *p = buf;
      size_t len = strlen(rtn->u.str);
      if (len > sizeof(buf)) {
	if ((p = malloc((ALLOC_T) (2 * len))) == NULL)
	  return(free_it(rtn, FALSE));
      }
      len = escape_url(rtn->u.str, p);
      (void) free_it(rtn, FALSE);
      rtn->what      = r_string;
      rtn->u.str     = strdup(p);
      rtn->size      = len;
      rtn->allocated = TRUE;
      if (p != buf)
	free((FREEPTR *) p);
    }
    break;
    

  case FX_OP_UNESCAPE : /* reverse operation */
    if(!form->u.f.args) 
      return(FALSE);
    if(!display_node_exec(after,fd,form->u.f.args,level,pos,np,attr,rsp,rtn) ||
       rtn->what != r_string ||
       rtn->u.str == NULL)
      return(free_it(rtn, FALSE));
    {
      char  buf[16384];
      char *p = buf;
      size_t len = strlen(rtn->u.str);
      if (len > sizeof(buf)) {
	if ((p = strdup(rtn->u.str)) == NULL)
	  return(free_it(rtn, FALSE));
      }
      (void) free_it(rtn, FALSE);
      len = unescape_url(rtn->u.str, p);
      rtn->what      = r_string;
      rtn->u.str     = strdup(p);
      rtn->size      = len;
      rtn->allocated = TRUE;
      if (p != buf)
	free((FREEPTR *) p);
    }

  }  /* switch form->u.f.func  */
  return(TRUE);
}

/*  */
static inline int display_node_var(after, fd, form, level, pos, np, attr, rsp, rtn)
 int after;
 FILE *fd;
 formxentry *form;
 int level;
 int pos;
 infonode *np;
 int *attr;
 reference_structure *rsp;
 formxret *rtn;
{
  switch(form->u.var) {
  case root:
    rtn->what = r_string;
    rtn->size = strlen(context);
    rtn->allocated = TRUE;
    if (!(rtn->u.str = malloc(rtn->size + 30))) 
      fatal_error("variable context: malloc failed\n");
    strcpy(rtn->u.str, context);
    break;
    
  case relpath : 
    rtn->allocated = TRUE;
    rtn->size = 30;
    rtn->what = r_string;
    rtn->u.str = NULL;
    (void)make_path(NULL, np, rtn->size, &rtn->u.str);
    break;
    
  case path : 
    rtn->allocated = TRUE;
    rtn->size = strlen(context) + 30;
    rtn->what = r_string;
    rtn->u.str = NULL;
    (void)make_path(context, np, rtn->size, &rtn->u.str);
    break;
    
  case self : 
    rtn->what = r_binary;
    rtn->allocated = FALSE;
    rtn->u.str = np->val.attvalue;
    rtn->size = np->val.attsize;
    break;
    
  case gpd : 
    rtn->allocated = FALSE;
    rtn->what = r_string;
    rtn->u.str = rsp->template->idname;
    break;
    
  case gpd_expand : 
    rtn->what = r_string;
    rtn->u.str = findexpandtag(np);
    if (rtn->u.str == NULL) {
      char *gpd_db = NULL;
      /* notice that gpd_db sets to non-null if allocated in NODE */
      rtn->u.str = find_default_gpd(&np->obj,&gpd_db);
      rtn->allocated  = gpd_db != NULL;
      rtn->size  = gpd_db != NULL ? strlen(gpd_db) : 0;
    } else {
      rtn->allocated = TRUE;
      rtn->size = strlen(rtn->u.str);
    }
    break;
    
  case field : 
    rtn->allocated = FALSE;
    rtn->what = r_string;
    rtn->u.str = np->field;
    break;
    
  case group : 
    rtn->allocated = FALSE;
    rtn->what = r_string;
    rtn->u.str = np->group;
    break;
    
  case position :
    rtn->allocated = FALSE;
    rtn->what = r_integer;
    rtn->u.value = pos;
    break;
    
  case clevel : 
    rtn->allocated = FALSE;
    rtn->what = r_integer;
    rtn->u.value = level;
    break;

  case ctype : 
    rtn->allocated = FALSE;
    rtn->what = r_integer;
    rtn->u.value = np->index;
    break;
    
  case size : 
    rtn->allocated = FALSE;
    rtn->what = r_integer;
    rtn->u.value = np->val.attsize;
    break;
    
  case line : 
    { 
      int  i  = np->val.attsize;
      char *p = np->val.attvalue;

      rtn->allocated = FALSE;
      rtn->what = r_integer;

      for(rtn->u.value = 0; i-- >0; )
	if(*p++ == '\n') rtn->u.value++;
      break;
    }
  } /* switch on variable type */
  return(TRUE);
}

/*  */
/*********************************************************************
 * Function: static int display_node_exec(after, fd, form, level, pos, np, attr, rsp, rtn)
 *
 *                 int  after  false if in pre phase, true in post phase
 *                FILE *fd     output file
 *          formxentry *form   current xformat to execute
 *                 int  level  downard positon when doing children
 *                 int  pos    position when doing siblings
 *            infonode *np     where we find the xformats 'code'
 *                 int *attr   current set of accumalted attributes
 * reference_structure *rsp    current reference structure 
 *            formxret *rtn    return value ...
 *
 * The entry point for our interpreter of XFORMATS structure. Return TRUE
 * if OK, else FALSE. Return value is in rtn, the caller is responisble for
 * free any memory allocated.
 *
 * Modifications:
 *	<list mods here with name and date>
 */ 
static int display_node_exec(after, fd, form, level, pos, np, attr, rsp, rtn)
 int after;
 FILE *fd;
 formxentry *form;
 int level;
 int pos;
 infonode *np;
 int *attr;
 reference_structure *rsp;
 formxret *rtn;
{
  rtn->allocated = FALSE;
  rtn->size      = 0;
  rtn->what      = r_fail;

  if (form == NULL)
    return(TRUE);

  switch (form->what) {
  case seq : /* sequence operation */
    do {
      display_node_exec(after,fd,form->next,level,pos,np,attr,rsp,rtn);
      form = form->next;
    } while(form != (formxentry *) NULL);
    break;

  case attribute : /* if true, attribute restrict current level */
    rtn->what = r_integer;
    rtn->u.value = 
      display_node_attr(after, level, pos, np, *attr | form->u.attr, rsp);
    break;

  case identifier :
  case string :
    rtn->what   = r_string;
    rtn->u.str  = form->u.str;
    break;

  case integer :
    rtn->what = r_integer;
    rtn->u.value = form->u.value;
    break;

  case function : 
    return(display_node_func(after, fd, form, level, pos, np, attr, rsp, rtn));

  case variable : 
    return(display_node_var(after, fd, form, level, pos, np, attr, rsp, rtn));

  case format : /* "old" style string with attribute etc */
    { int a = *attr | form->u.form.attr;

      /* mm, if we have once marked somewhere, we might have printed this  */
      if (np->printed) break;
	
      /* if true, attribute restrict current level */
      if(display_node_attr(after, level, pos, np, a, rsp)) {
	rtn->what = r_integer;
	rtn->u.value  = FALSE;
	break;
      }

      /* and print it ... */
      if (form->u.form.val != NULL) {
	int freeit = FALSE;
	char *p;

	p = a & FX_ATTRNOREMAP ? zerolen(np->val.attvalue) :
	  remap_string(zerolen(np->val.attvalue), NULL, &freeit);

	fprintf( fd, form->u.form.val, p);

	/* if we have allocated new memory, free it */
	if (freeit) free((FREEPTR *) p);
	/* set printed flag *only* if we have ONCE enabled */
	if(a & FX_ATTRONCE)  np->printed = 1;

      } /* form->u.form.val != NULL */
      break;
    } /* case format: */
  } /* switch -> what */
  return (TRUE);
}

/*  */
/**********************************************************************
 * Function: static void display_node(fd, rsp, np, level, pos, attr)
 *
 *                 FILE *fd     output file descriptor
 *  reference_structure *rsp    reference structure 
 *             infonode *np     infonode (where we are)
 *                  int  level  increment by one when following children
 *                  int  pos    increment by one when following sibling
 *                  int  attr   current accumulated attribute 
 *
 * This function builds a "display" for an infonode and prints output
 * on fd. 
 *
 * Modified:
 * 	<list mods here with name and date>
 */
static void display_node(fd, rsp, np, level, pos, attr)
  FILE *fd;
  reference_structure *rsp;
  infonode *np;
  int level;
  int pos;
  int attr;
{
  formxret rtn;
  identry *idp;
  int current_stack_level = format_stack_ptr; /* used to restore */

  /* add the inherited formatting attributes to stack. */
  stack_formats(np->head->tmpl->value);		/* Generic formatting */
  stack_formats(np);				/* Specific formatting */

   /* ---------------------------------------------------------------------- */
  if (!np->printed) {
    int current = current_stack_level;

    (void) display_node_exec(FALSE, fd, np->xformats, 
			     level, pos, np, &attr, rsp, &rtn);
    (void) free_it(&rtn, FALSE);
    (void) display_node_exec(FALSE, fd, np->head->tmpl->value->xformats,
			     level, pos, np, &attr, rsp, &rtn);
    (void) free_it(&rtn, FALSE);

    /* notice that we have already done the inherited stuff on this level */
    for(; --current > 0; current-- ) {
      (void) display_node_exec(FALSE, fd, format_stack[current], 
			       level, pos, np, &attr, rsp, &rtn);
      (void) free_it(&rtn, FALSE);
    }
#if 0
    /* this is default ... blagh, need to check for placeholders, too */
    if(!done && np->subnodes == NULL && np->val.attvalue != NULL)
      if (!np->notindb || (attr & FX_ATTRPH)) {
	fprintf( fd,"\n%s", zerolen(np->val.attvalue));
        np->printed = 1; /* mark as printed */
      }
#endif
  }

  /* ---------------------------------------------------------------------- */
  if ((idp = np->subnodes) != (identry *) NULL) {
    infonode *copy = np;
    int copypos = pos;
    while (idp != (identry *) NULL) {
      for (np=idp->value, pos=0; np != (infonode *) NULL; np=np->next, pos++) {
	display_node(fd, rsp, np, level+1, pos, attr);
      }
      idp = idp->next;
    }
   np = copy;
   pos = copypos;
  }

  /* ---------------------------------------------------------------------- */
  /* take care of 'AFTER' output stuff */
  {
    int current = current_stack_level;

    (void) display_node_exec(TRUE, fd, np->xformats, 
			     level, pos, np, &attr, rsp, &rtn);
    (void) free_it(&rtn, FALSE);
    (void) display_node_exec(TRUE, fd, np->head->tmpl->value->xformats, 
			     level, pos, np, &attr, rsp, &rtn);
    (void) free_it(&rtn, FALSE);
    for(; --current > 0; current-- ) {
      display_node_exec(TRUE, fd, format_stack[current],
			level, pos, np, &attr, rsp, &rtn);
      (void) free_it(&rtn, FALSE);
    }
  }

  /* ---------------------------------------------------------------------- */
  /*  reset the format stack to level when entered this function */
  format_stack_ptr = current_stack_level;
}

/*  */
/*********************************************************************
 * Function: static void stack_formats(infonode *np)
 *
 * This function "stacks" formatting attributes for a particular 
 * infonode onto a global stack which will then be copied and sent
 * to the propagation queue processing routine.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static void stack_formats(np)
  infonode *np;
{
  formxentry *fp = np->xformats;

  while(fp != (formxentry *) NULL) {
    formxentry *p;
    if (fp->what == seq) {
      p = fp->next;			/* to check */
      fp = p;
    } else {
      p = fp;				/* check this */
      fp = NULL;			/* and we are done */
    }
    
    if (p->what == format) 
      if ((p->u.form.attr & FX_ATTRINHERIT) == FX_ATTRINHERIT) {
	format_stack[format_stack_ptr++] = p;
	if (format_stack_ptr >= NUM_FORMATS) 
	  fatal_error("stack_formats: overflow in num of formatting el.\n");
      }
  } /* while */
}

/*  */
/**********************************************************************
 * Function: static int firsttagname(void *extra, char *tag, int count, int length)
 *
 * A mapfunction called to map over all fieldnames in a linkgroup
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int firsttagname(extra, tag, count, length)
  void *extra;
  char *tag;
  int count, length;
{
  char **p = (char **) extra;
  if (count != 0)
    *p = strdup(tag);
    return (0);
}

/*  */
/*********************************************************************
 * Function: find_default_gpd(label *gotit, char **gpdname_db);
 *
 * Try to find a gpd using the following 
 *
 *  1 - check if we have a LOGIN attr name (in fact WEBROOT can have that
 *      since gotit may point to WEBROOT ... )
 *  2 - if not - check if userroot => default to USER_GPD
 *  3 - else user DEFAULT_GPD
 *
 *  Notice that gpdname_db is non-null when allocated i NODE, else
 *  if it a constant, it is NULL.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static char * find_default_gpd(gotit,gpdname_db)
 label *gotit;
 char **gpdname_db;
{
  *gpdname_db = NULL;
  if (GLN_GETLINKNAMES(gotit, AIM_LOGIN, firsttagname, 
		       (void *)gpdname_db) == SUCCESS 
      && *gpdname_db != NULL)
    return(*gpdname_db);
  else if (our_userroot.vs == gotit->vs && our_userroot.inst == gotit->inst) 
    return(USER_GPD);
  else 
    /* TODO better way to find gpd name .. when traversing the structure ? */
    return(DEFAULT_GPD);
}

/*  */
/*********************************************************************
 * Function: void reset_macros(char *mode)
 *
 * resets load the macro definitions ....
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
void reset_macros(mode)
  char *mode;
{
  export_mode *exp = find_export_mode(mode);

  if(exp->macros != NULL) {
    int i;
    for(i=0; i < exp->macrosize; i++) 
      free_xformats(exp->macros[i].code);
    free((FREEPTR *) exp->macros);
    exp->macros = NULL;
    exp->macrosize = 0;
  }
}

/*  */
/*********************************************************************
 * Function: int load_macros(char *mode, char *macros)
 *
 * load the macro definitions, where the macros string looks like
 * this:
 *
 * <name1>\0<macrocode1>\0 
 * ....
 * <namen>\0<macrocoden>\0\0
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
int load_macros(mode, macros)
  char *mode;
  char *macros;
{
  export_mode *exp;
  int cnt = 0;

  /* find or create export_mode node */
  exp  = find_export_mode(mode);

  /* count number of entries to add (in buf) */
  {
    char *p = macros;
    for(cnt=0; *p != '\0'; cnt++) {
      while(*p++ != '\0');
      while(*p++ != '\0');
    }
  }

  /* add extra space to main array ... */
  if (exp->macros ==  NULL) 
    exp->macros = malloc(sizeof(t_macro) *  ++cnt);
  else
    exp->macros = realloc(exp->macros, sizeof(t_macro) * (cnt + exp->macrosize));

  if (exp->macros == NULL)
    fatal_error("load_macros: realloc/malloc failed");

  /* ok, let add those entries ... */
  {
    char    *p  = macros;
    int      pos = exp->macrosize == 0 ? 0 : exp->macrosize - 1;
    infonode inp;			/* for hanging up xformats on */

    exp->macrosize += cnt;

    for(; pos < exp->macrosize-1 && *p != '\0'; pos++) {
      exp->macros[pos].name = p;	/* save name */
      while(*p++ != '\0') ;		/* skip until code */
      inp.xformats = NULL;		/* default if failure */
      (void) DoParseExecInfoNode(&inp, p, AIM_FORMATX);
      exp->macros[pos].code = inp.xformats;
      while(*p++ != '\0') ;		/* skip until name */
    } /* for */
  }
  return(exp->macrosize);
}

/*  */
/*********************************************************************
 * Function: static t_macro * find_macro(char *name)
 *
 * Search (linear - still soo small, will make into a binary search later)
 * through for a matching macro name
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static t_macro * find_macro(name)
 char *name;
{
  int i;
  t_macro *macros = current_mode->macros;

  for(i=0; i < current_mode->macrosize; i++, macros++) 
    if(strcmp(name,macros->name) == 0)
      return(macros);

  return(NULL);
}

/*  */
/*********************************************************************
 * Function: void reset_shortname(char *mode)
 *
 * resets load the short name mapping array
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
void reset_shortname(mode)
  char *mode;
{
  export_mode *exp = find_export_mode(mode);
  if(exp->shortpath != NULL)
    free((FREEPTR *) exp->shortpath);
  exp->shortpath = NULL;
}

/*  */
/*********************************************************************
 * Function: void load_shortname(char *mode, char *buf)
 *
 * load the short name mapping array from the area (string) pointed
 * to by buf. Format should be:
 *
 * shortname\ngroup\nfield\n ...
 *
 * Notice that buf is modified ('\0' are inserted instead of \n), 
 * and thus you should either make a copy or allow writable strings.
 * If *arrp is not NULL, we use realloce to extend the area.
 *
 * Return number of entries in the array after adding the new one.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
int load_shortname(mode, buf)
  char *mode;
  char *buf;
{
  export_mode *exp;
  int count;
  int oldsize;
  t_shortpath **arrp;

  /* find or create export_mode node */
  exp  = find_export_mode(mode);
  arrp = &(exp->shortpath);
  oldsize = exp->shortsize;

  /* count number of entries to add (in buf) */
  {
    char *p = buf;
    count   = 0;
    while(*p != '\0')
      if (*p++ == '\n') count++;
    count = count/3;
  }

  /* add extra space to main array ... */
  if ((*arrp) ==  NULL) {
    count++;
    (*arrp) = malloc(sizeof(t_shortpath) *  count);
  }
  else
    (*arrp) = realloc((*arrp), sizeof(t_shortpath) * (count + oldsize));

  if ((*arrp) == NULL)
    fatal_error("load_shortname: realloc/malloc failed");

  /* ok, let add the entries ... */
  {
    char *p = buf;
    int  pos = oldsize == 0 ? 0 : oldsize - 1;

    while(p != NULL && *p != '\0') {
      char *name  = p;
      char *group = NULL;
      char *field = NULL;

      if((group = strchr(name,'\n')) == NULL ||
	 (field = strchr(group+1,'\n')) == NULL) {
	(void) fprintf(stderr,"load_shortname: funny format on input buffer");
	free((FREEPTR *) (*arrp));
	(*arrp) = NULL;
	return(0);
      } else {
	/* chop off all found fields, notice special case of last char */
	 if((p = strchr(field+1,'\n')) != NULL) 
	   *p++ = '\0';
	 *group++ = '\0';
	 *field++ = '\0';
	 
	 /* add to array and insert in right place */
	 (*arrp)[pos].name = name;
	 (*arrp)[pos].group = group;
	 (*arrp)[pos].field = field;
	 pos++;
       } /* strchr .. */
    } /* while */
    (*arrp)[pos].name = (*arrp)[pos].group = (*arrp)[pos].field = NULL;
  }
  return(exp->shortsize += count);
}

/*  */
/*********************************************************************
 * Function: static void find_shortname(infonode *np, char **group, field)
 *
 * Search (linear - still soo small, will make into a binary search later)
 * through for a matchin np->group & np->field, and updated *field (to NULL)
 * and *group (to found short name). If no match, we do not change either
 * of *group nor *field.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static void find_shortname(np, group, field)
 infonode *np;
 char **group;
 char **field;
{
  int i;
  t_shortpath *shortpath = current_mode->shortpath;

  /* notice that the last pair i pt to NULL .. */
  for(i=0; i < current_mode->shortsize-1; i++) 
    if(strcmp(np->group,shortpath[i].group) == 0 &&
       strcmp(np->field,shortpath[i].field) == 0) {
      *group = shortpath[i].name;
      *field = NULL;
      return;
    }
}

/*  */
/*********************************************************************
 * Function: static int get_fullname(char *part, char **group, field, int pos)
 *
 * Reverse function of find_shortname, e.g. given a shortname, in *group,
 * update *group, *field with the full name. Notice that *part points
 * to remaints, which might be a position, and thus we update *pos, to.
 *
 * Return - if we found a successfull match LINKITEM, else IMAGE. 
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static int get_fullname(part, group, field, pos)
 char *part;				/* points to position string */
 char **group;				/* points to shortname ... */
 char **field;
 int  *pos;
{
  int i;
  t_shortpath *shortpath = current_mode->shortpath;

  for(i=0; i < current_mode->shortsize-1; i++) 
    if(strcmp(*group,shortpath[i].name) == 0) {
      *group = shortpath[i].group;
      *field = shortpath[i].field;
      *pos = atoi(part);
      return(LINKITEM);
    }
  return(IMAGE);
}

/*  */
/*********************************************************************
 * Function: static int get_link(int doremap, char *part, char **group, field, 
 *                               int *pos, char **gpd)
 *
 * Break down `part' (created by printf_part) into group/field/pos
 * and possible gpd. This the overall syntax - notice that ':','-'
 * are defined as constant (SEP_ITEM/SEP_GPD):
 *
 * We should the following possible contents of each parts 
 *
 * GROUP:FIELD:POS-GPD   => where ':' are define somewhere above
 * GROUP:FIELD:-GPD      => expand all link values as a folder ..
 * ShortName:[0-9]+-GPD   => expands into GROUP FIELD 
 * GROUP-GPD             => look for a matching image among link fields.
 *
 * notice that -GPD is optional and might not be present
 * 
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static int get_link(doremap, part, group, field, pos, gpd)
 int   doremap;
 char *part;
 char **group;
 char **field;
 int  *pos;
 char **gpd;
{
  char *q;

  /* nice defaults */
  *group = *field = *gpd = NULL;
  *pos = 1;				

  /* start looking throught the part for - or ':' */
  if (*part == '\0')   return(-1);	
  *group = part;			/* is first */
  for(; *part != SEP_ITEM; part++)
    if (*part == SEP_GPD) {		/* GROUP-GPD */
      *part++ = '\0';			/* chop of '-' */
      *gpd = part;
      return(IMAGE);
    } else if (*part == '\0')		/* GROUP  */
      return(IMAGE);
  *part++ = '\0';			/* chop off ':' */

  /* do we have any field part? */
  if (*part == '\0') return(IMAGE);			
  *field = part;
  for(; *part != SEP_ITEM; part++)		
    if (*part == SEP_GPD) {		/* shortname:pos-GPD */
      *part++ = '\0';			/* chop of '-' */
      *gpd = part;
      if (doremap)
	return(get_fullname(*field, group, field, pos));
      else {
	*pos = atoi(*field);		/* convert pos */
	*field = NULL;			/* no field */
	return(LINKITEM);
      }
    } else if (*part == '\0') {		/* Shortname:Pos  */
      if (doremap)
	return(get_fullname(*field, group, field, pos));
      else {
	*pos = atoi(*field);		/* convert pos */
	*field = NULL;			/* no field */
	return(LINKITEM);
      }
    }
  *part++ = '\0';			/* chop off ':' */

  /* do we have any position part */
  if (*part == '\0') return(LINKGROUPTAG); /* GROUP:FIELD: */
  q = part;
  for(; *part != '\0'; part++)		/* default to pos 1 ... */
    if (*part == SEP_GPD) {		/* GROUP:FIELD:POS-GPD */
      *part++ = '\0';			/* chop of '-' */
      *gpd = part;
      break;
    } 

  *pos = atoi(q);
  return(LINKITEM);
}

/*  */
/*********************************************************************
 * Function: static int follow_path(char *path, label *gotit, 
 *                                  char **prefix, **gpdname,
 *                                  FILE *logdfd);
 *
 * Reverse operation of make_path, break down the path (=url) into
 * its parts, and traverse the ACCESS path into the database. Returns
 * the found object in gotit, as well as the username, possible prefix
 * to the object, as well as the gpd to expand the object (gotit) with.
 *
 * Return: FALSE if failed to follow path and TRUE if found object.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
static int follow_path(path, gotit, prefix, gpdname, logfd)
 char *path;
 label *gotit;
 char **prefix;
 char **gpdname;
 FILE *logfd;
{
  int i;
  int slash_at_end;
  label next;
  int cnt = 0;
  
  char *p = path;	
  char **parts = NULL;

  /* first, count number of slashes to build an array to keep them in */
  while (*p != '\0') if (*p++ == '/') cnt++;

  /* notice that an ending slash => prefix starts with ../ and not ./ */
  if((slash_at_end = p[-1] == '/')) {
    p[-1] ='\0';			  /* see comment below */
    cnt--;				  /* remove slash  */
  } 

  if ((parts = (char **) malloc((ALLOC_T) sizeof(char *) * (cnt + 2))) == NULL) {
    (void) fprintf(logfd, 
		   "follow_path: failed to allocate memory (%d)\n", 
		   cnt);
    return(FALSE);
  }
  
  /* fill it up with the different parts and terminate strings ... */
  /* should not start with a '/' and notice that it do not end with a '/' */
  cnt=0; parts[cnt++]=path;
  for( p=path; *p != '\0'; p++)
    if (*p == '/') {
      *p++ = '\0';
      parts[cnt++]=p;
    }
  parts[cnt]=NULL;

  /* allocate memory for prefix before we chops up the parts in get_link */
  { size_t partslen    = strlen(parts[cnt-1]) + 5;

    if ((*prefix = malloc((ALLOC_T) partslen)) == NULL) {
      (void) fprintf(logfd,"malloc failed when building prefix\n");
      free((FREEPTR *) parts);
      return(FALSE);
    }
    strcpy(*prefix, slash_at_end ? "../" : "./");
    strcat(*prefix, parts[cnt-1]);	/* and save a copy of the last part */
  }

  /* mm, remove any GPD specification at end ... */
  {
    char *group, *field;
    int pos;
    if(get_link(TRUE, parts[0], &group, &field, &pos, gpdname) != IMAGE) {
      (void) fprintf(logfd, 
		     "Cannot handle `%s' as initial path\n",
		     parts[0]);
      free((FREEPTR *) parts);      
      return(FALSE);
    } 
    /* first, the username maps into a userroot ... */
    if (GLI_GETLINKITEM(&sysroot, AIM_USERS, parts[0], 1, &our_userroot)) {
      (void)fprintf(logfd, "Cannot retrieve user `%s' root.\n", parts[0]);
      free((FREEPTR *) parts);
      return(FALSE);
    }
  }

  /* second, check for WWW_ROOT to point to webroot, 
   * if not set gotit with current user root */
  if (GLI_GETLINKITEM(&our_userroot, WWW_ROOT, 1, gotit)) 
    copylbl(gotit, &our_userroot);		

  /* now, we should the following possible contents of each parts  */
  for(i=1; i < cnt; i++) {
    char *group, *field;
    int pos;
    switch(get_link(TRUE, parts[i], &group, &field, &pos, gpdname)) {
       case LINKITEM :
	 if (GLI_GETLINKITEM(gotit, group, field, pos, &next) == SUCCESS)
	   break;
         (void)fprintf(logfd, "\nFailed to follow `%s':`%s':%d\n ",
		       group, field, pos);
       case IMAGE :        /* look for matching in ???? */
       case LINKFIELDTAG : /* missing position ... */
       case LINKGROUPTAG : /* missing field and position */
       default:
	free((FREEPTR *) parts);
	return(FALSE);
    } /* switch */
    copylbl(gotit,&next);
  }
  free((FREEPTR *) parts);

  /* build a context with a `./'`../' only if not at userroot */
  /* since we allow : in paths we need to prefix with a / */
  {
    char *group, *field, *gpdname;
    int pos;    
    int offset = slash_at_end ? 3 : 2;
    /* get rid of any gpd, reason behind this stuff */
    if (get_link(FALSE,*prefix+offset, &group, &field, &pos, &gpdname) == LINKITEM) 
      (void) printf_part(*prefix+offset, group, field, pos);
    else 
      (void) printf_part(*prefix+offset, group, field, 0);
  } 

  return(TRUE);
}

/*  */
/*********************************************************************
 * Function:  void load_html()
 * 
 * Load HTML predefined mappings if not done !
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 */
void load_html()
{
  static int loaded = FALSE;
  /* http://info.cern.ch/hypertext/WWW/MarkUp/Entities.html */
  /* static since we need this around */
  static char *html_remap = "\001<&lt;\0\001>&gt;\0\001&&amp;\0\001\"&quot;\0\000";

  /* macros - static since we have pointers into this ...*/
  /* TODO - add <LINK REV="made" hef="mailto:%s"> */
  /*      - add epilong name + link              */
  /*      - get ATTR SYSTEM:Created < ... > => */
  /*      - name : USERRoot + SYSTEM:WWWRoot + ATTR WWW:Username */
  /*      - adr  : USERRoot + SYSTEM:WWWRoot + ATTR WWW:"E-mail address" */
  static char *html_macros = "ROOTBEFORE\0\
\"\\\\n<HTML><HEAD><TITLE>%s</TITLE>\\\\n</HEAD><BODY>\"(BEFORE)\0\
ROOTAFTER\0\
\"\\\\n</BODY></HTML>\"(AFTER)\0\000";
 
  /* shortname mappings */   
  char *html_shortpath = "TP\nTEXT\nParagraph\nTT\nTEXT\nTitle\n\
TS\nTEXT\nSublevel\nTA\nTEXT\nAbstract\n\
TO\nTEXT\nOwner\nTp\nTEXT\nparagraph\n\
Tt\nTEXT\ntitle\nTs\nTEXT\nsublevel\n\
DA\nDOCUMENT\nAuthor\n\
UR\nUSER\nRoot\n\
FI\nFOLDER\nItem\n\
FF\nFOLDER\nItem.from\n\
SG\nSYSTEM\nGPDmaplist\n";

  if(loaded)
    return;

  /* set current mode to be HTML */
  current_mode = find_export_mode(HTML);
  if (current_mode->initstring == (char *) NULL)
    current_mode->initstring = "Content-Type: text/html\n\n";

  /* since load_shortname chops things up ... */
  (void) load_shortname(HTML, strdup(html_shortpath));

  /* load_charset keeps pointers *into* the string */
  load_charset(HTML, html_remap);

  /* load macros */
  (void) load_macros(HTML, html_macros);

  loaded = TRUE;
}
 
/*  */
/*********************************************************************
 * Function:  void create_html(int sockfd, FILE *logfd, long key)
 * 
 * Read path/url from fd which is connect to addr using logfd for
 * logging on.
 *
 * Modifications:
 *	<list mods here with name and date>
 * 
 *
 * TODO: - export + 'accepted' stuff from cgi2lincks 
 *                + parse here ..
 *                + make it read "real" HTTP header ..
 *       - path parsing -> 
 *                + first parse/split
 *                + then un escape
 *       - prolog + add <link rev=make>
 *                + return info concerning last change
 *       - epilog + add username, link to e-mail, last changed
 *       - add function/macro for construct paths
 *       - add something to handle next/prev/first/last/up
 *              + add macros for `button' for n/p/f/l at top and end if ..
 *       - GET = (curret) HEAD -> special xformat, latex -> special xformats 
 *       - look into placeholder ... ->flag & ...
 *       - GPD  + change syntax into ;gpd=value ...
 *              + use special 'gpd' to expand and following link to place
 *                => default gpd at end!
 *              + add FLAT -> check for FLAT/DEEP 
 *              + make the GPD parsing stuff in get_link/follow_path 
 *                independent of HTML since::
 *                * get_link *knows* that a GPD follows  '-' ..
 *                * follow_path using WWW_ROOT
 *       - GPD restrict set to those hanging of `www user root'
 *       - PEN - macro function or in header?/object?
 *       - FORMS => generate/reparse & state(?)
 *         + cgi2lincks
 *           * add support for head(put/post) ?
 *       - how is setuid/username really used ?
 *         + passwd/validate/setuid @ dbs/ws => how ? 
 *         + gpdmaplist (findtag) checking to use our root ?
 *       - macro       + use binary search
 *       - short name  + use binary search 
 *
 * AIM   => move constant into header file ("user"/LOGIN etc)
 *
 * EXTERNAL to this module: + macro storage/retrieve(export)
 *                          + charset storage/retrieve(export)
 *                          + shortname storage/retrieve(export)
 *                          + configuration file
 *                          + callback(?) or when new export? 
 *
 */
void create_html(sockfd, logfd, key)
  int   sockfd;
  FILE *logfd;
  long  key;
{
  label gotit;				/* final object */
  FILE *wrfd;				/* socket to write on */
  char  *Context = NULL;
  char  *scriptname  = NULL;		/* script name from cgi2lincks */
  char  *gpdname = NULL;		/* gpd name or NULL */
  char  *gpdname_db = NULL;		/* when getting via DB */
#ifdef THESE_ARE_NOT_USED
  attrval *username;			/* store name from WWW_ROOT */
  attrval *emailadr;			/* store e-mail address from WWW_ROOT */
#endif /* THESE_ARE_NOT_USED */
  char  url[4096];			/* buffer for url */
  char  inurl[4096];			/* buffer for unescaped url */
  char *purl=url;			/* pts to url */
  int   len_loginuser = strlen(LL_loginuser);
  int   len;
  char  *p;

  reference_structure *rsp = NULL; 

  if ((len = read(sockfd, url, sizeof(url))) < 0) {
    (void) fprintf(logfd, 
		   "create_html: failed to read from socket - %s",
		   sys_errlist[(errno > sys_nerr) ? 0 : errno]);
    close(sockfd);
    return;
  }

  /*  chop of any trailing cr/lf */
  while(len > 0 && (url[len-1]=='\012' || url[len-1] =='\015'))
    url[--len]='\0';

  if ((wrfd = fdopen(sockfd,"r+")) == NULL) {
    (void) fprintf(logfd,
		   "create_html: failed to fdopen on socket stream - %s\n",
		   sys_errlist[(errno > sys_nerr) ? 0 : errno]);
    close(sockfd);
    return;
  }

  /*
   * we are supposed to get something which looks like
   *
   * <key>/<scripname>/<path>
   *
   * where currently <loginuser> should be the first part
   * of the path ... also, notice the following:
   *
   * 1) if key  => 0 no need for key initially ...
   * 2) scriptname *must* be the initial part when we return path
   * 3) if path is ended by a '/' we cannot simply use `./' first
   *    on path, but maybe (?) `../'. We have a problem since we
   *    allow `:' in parts ... and thus needs to add a something 
   *    before
   *
   */
  if (key != 0) {
    long in_key = strtol(url, &purl, 10);
    if (in_key != key) {
      (void) fprintf(logfd,
		     "create_html: key %ld given instead of %ld\n",
		     in_key, key);
      fclose(wrfd);
      return;
    };
    purl++;				  /* move past the ending / */
  } /* key != 0 */

  /* get the end of the scriptname and chop off it */
  if ((purl = strchr(scriptname = purl,'/')) != NULL) 
    *purl++ = '\0';			  

  /* get the to end of the username part of the URL */
  for(p = purl; *p != '\0' && *p != '/' && *p != SEP_GPD; p++);
  if (p == purl) {
    (void) fprintf(wrfd,"Content-type: text/plain%c%c",10,10); 
    (void) fprintf(wrfd, "No user specified in URL!\n");
    fclose(wrfd);
    return;
  }

  /* TODO ... validate the usernname from the cgi2lincks script */
  if (p - purl != len_loginuser ||
      strncmp(purl, LL_loginuser, len_loginuser) != 0) {
    (void) fprintf(wrfd,"Content-type: text/plain%c%c",10,10); 
    (void) fprintf(wrfd, 
"Currently you cannot login/browse as user `%.*s'\n\n\
Only user `%s' is supported via this database client.\n",
		   (int) (p - purl), purl, LL_loginuser);
    fclose(wrfd);
    return;
  }

  (void) fprintf(logfd,"\nURL:`%s' ", purl);	

  /* get back spaces etc */
  (void) unescape_url(purl, inurl);

  /* and set current mode to be HTML before calling these routine .. */
  current_mode = find_export_mode(HTML);

  /* parse URL -> find object + gpd(?) */
  if (follow_path(inurl, &gotit, &Context, &gpdname, logfd)) {

    /* if no gpd given ... find one */
    if (gpdname == NULL) 
      gpdname = find_default_gpd(&gotit, &gpdname_db);
      
    (void) fprintf(logfd, "=> scriptname `%s' gpd `%s' obj %d %d\n",
		   scriptname, gpdname , gotit.vs, gotit.inst);

    /* build the reference structure */
    rsp = buildref(&gotit, gpdname, (reference_structure *)NULL);
    xblddisplay(rsp, wrfd, Context, HTML);
#if WWW_UPDATE
    /* TODO Here we'll release the objects now and then ... when ?? */
#endif
    free_ref(rsp);			  
  }
  else {
    (void) fprintf(wrfd,"Content-type: text/plain%c%c",10,10); 
    (void) fprintf(wrfd, "Failed to parse input URL:\n`%s'\n", inurl);
  }

  /* free up used stuff  ... */
  if (gpdname_db) 
    free((FREEPTR *) gpdname_db);
  if (Context) 
    free((FREEPTR *) Context);
  fclose(wrfd);
}

/*  */
/* borrowed from NCA httpd, util.c */
/*
 * str.c: string utility things
 * 
 * 3/21/93 Rob McCool
 * 
 */
static char x2c(what) 
  char *what;
{
    register char digit;

    digit = ((what[0] >= 'A') ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
    digit *= 16;
    digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
    return(digit);
}

static int unescape_url(url, out) 
  char *url;
  char *out;
{
  register int x,y;

  for(x=0,y=0;url[y];++x,++y) {
    if((out[x] = url[y]) == '%') {
      out[x] = x2c(&url[y+1]);
      y+=2;
    }
  }
  out[x] = '\0';
  return(x);
}

static int ind(s, c) 
 char *s;
 int c;
{
    register int x;

    for(x=0;s[x];x++)
        if(s[x] == (char) c) return x;

    return -1;
}

#define c2x(what,where) sprintf(where,"%%%2x",what)

static int escape_url(url,out) 
  char *url;
  char *out;
{
  register int x,y;

  for(x=0,y=0;url[x];x++,y++) {
    if(ind("% ?+&",out[y] = url[x]) != -1) {
      c2x(url[x],&out[y]);
      y+=2;
    }
  }
  out[y] = '\0';
  return(y);
}

#else
void create_html(sockfd, logfd, key)
  int   sockfd;
  FILE *logfd;
  long  key;
{ return; }
void load_html() { return; }
#endif /* n XFORMAT */
