/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, 1994, 1995, 1996, 1997 Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME: 	usfunsadd.c
 *
 * SCCSINFO:		$Id: usfunsadd.c,v 1.4 1997/04/12 20:34:08 lincks Exp $
 *
 * ORIGINAL AUTHOR(S): Ralph Rnnquist,  18 Sep 1987
 *
 * MODIFICATIONS:
 *	<list mods with name and date>
 *
 * DESCRIPTION:
 *
 * This file is part of the liblincks library.
 * It contains the functions that application programs call
 * in order to access a LINCKS database.
 *
 * ADDITIONAL DOCUMENTATION:
 *
 * See the document, "Application Interface to the LINCKS
 * Database System through the LIBLINCKS library.  Programmers Manual,"
 * found at ~lincks/Text/Misc/Liblincks.manual/liblincks_man.tex
 *
 * IF YOU CHANGE THIS FILE:
 *	-- YOU MIGHT ALSO NEED TO UPDATE THE MANUAL.
 *	-- the externally-callable functions of this file are
 *		declared in liblincks.h, so you may need to change it.
 */
/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include "config.h"	/* includes system dependent includes */
#include "sunlincks.h"
#include "dbcodes.h"

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
#include "f_usfunsadd.h"

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
#include "f_sysfuns.h"

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern molecule *LL_moltbl;

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/

/*LINTLIBRARY*/

/*  */
/*********************************************************************
 * Function: errcode SET_LATEST(label *lbl,int clear)
 *
 * Add indicated version to list of latest. Clear list first if desired.
 *
 * Modifications:
 *      <list mods with name and date>
 */

errcode SET_LATEST(lbl,clear)	/****/
     label *lbl;
     int clear;
{
    linktag *latest = (linktag *)0;
    linkitem *p = (linkitem *)0;
    linkitem *q = (linkitem *)0;

    if (lbl->inst == INST_UNBOUND) {
	return 1; /* Error: label must be bound. */
    }
    if (LL_sysgetlink(lbl->vs, SYSTEM_LATEST, &latest) != 0) {
	return 2; /* Error accessing SYSTEM_LATEST */
    }

    if (clear) { /* Remove what is currently there, except transient */
	q = latest->value;
	latest->value = (linkitem *)0;
	while ((p = q) != (linkitem *)0) {
	    q = p->next_item;
	    if (LL_moltbl[p->link.inst].edited) {
		(void)LL_sysputlinkitem(lbl->vs,SYSTEM_LATEST, 1,&(p->link));
	    }
	    free((FREEPTR *)p);
	}
    }

    /* Add the new. */
    for (p = latest->value; p; p = p->next_item) {
	if (lbl->inst == p->link.inst)
	  return 0; /* Label is already in the list. */
    }
    (void)LL_sysputlinkitem(lbl->vs,SYSTEM_LATEST, 1,lbl);
    return 0;
}
