/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1994, 1995, 1996, 1997, Martin Sjlin
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * MODULE NAME:         cgi2lincks.c
 *
 * SCCSINFO:		$Id: cgi2lincks.c,v 1.4 1997/04/12 20:34:57 lincks Exp $
 *
 * ORIGINAL AUTHOR(S):  Martin Sjlin 1994-04-29
 *                      based upon code by Rob McCool
 *
 * DESCRIPTION:
 *
 * CGI routine for using together with LINCKS WWW gateway.
 * 
 *
 * MODIFICATIONS:
 *	<list mods here with name and date>
 */

/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include "config.h"
#include "machine.h"

#include "liblincks.h"			/* to get SUCCESS/FAIL */

#include <netdb.h>			/* to get struct hostent */
#ifdef  __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif /* n __STDC__ */

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNALLY-AVAILABLE DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern char *sys_errlist[];	/* errno.h */
extern int sys_nerr;		/* errno.h */
extern int errno;		/* errno.h */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
#include "f_tcp.h"

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
/* none */

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
void plustospace P_((char *str));

/*********************************************************************
 * INTERNAL (STATIC) DATA:
 *********************************************************************/
static char *prog = NULL;
static char *path_info   = NULL;
static char *user_name   = NULL;
static char *script_name = NULL;
static char *query       = NULL;
static char *server_name = "";
static char *our_name    = "cgi2lincks";

static char buf[16384];

static char *getmsg = 
"This script should be referenced with a METHOD of GET.\n\
If you don't understand this, see this \n\
<A HREF=\"http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/Docs/\
fill-out-forms/overview.html\">forms overview</A>";

/*  */
/**********************************************************************
 * Function: void die (FILE *fd, char *format, ...)
 *
 * error routine for ...
 *
 * Modifications:
 *      <list mods with name and date>
 */
#ifdef __STDC__
void die(FILE *fd, char *format, ...)
#else
void die(va_alist)
  va_dcl
#endif /* n __STDC__ */
{
  va_list ap;
#ifndef __STDC__
  FILE *fd;
  char *format;
  va_start(ap);
  fd     = va_arg(ap, FILE *);
  format = va_arg(ap, char *);
#else
  va_start(ap,format);
#endif
  (void)  fprintf(fd, "Content-Type: text/plain%c%c",10,10);
  (void)  fprintf(fd, 
		  "Contact the server adminstrator for %s\n\n", 
		  server_name);
  (void) vfprintf(fd, format, ap);
  exit(0);
}

/*  */
/**********************************************************************
 * Function: void connect_error (char *format, ....)
 *
 * callback error routine used for tcp_connect  ...
 *
 * Modifications:
 *      <list mods with name and date>
 */
#ifdef __STDC__
void connect_error(char *format, ...)
#else
void connect_error (va_alist)
  va_dcl
#endif /* n __STDC__ */
{
  va_list ap;
  char *p = buf;
#ifndef __STDC__
  char *format;
  va_start(ap);
  format = va_arg(ap, char *);
#else
  va_start(ap,format);
#endif
  strcpy(p,"Error during connection to the LINCKS database server:\n\n");
  p += strlen(p);
  vsprintf(p, format, ap);
}

/*  */
/**********************************************************************
 * Function: main(int argc, char *argv[]):
 *
 * error routine used for tcp_server ...
 *
 * Modifications:
 *      <list mods with name and date>
 */
int main(argc, argv)
  int argc;
  char *argv[];
{
  FILE *f;
  char *port, *key = NULL, *last = NULL;
  char host[4096];

  server_name  = getenv("SERVER_NAME");
  path_info   = getenv("PATH_INFO");
  user_name    = getenv("REMOTE_USER");
  script_name = getenv("SCRIPT_NAME");
  if ((query  = getenv("QUERY_STRING")) == NULL) 
    query = "";
  plustospace(query);

  /* get a nice program name */
  if ((prog = strrchr(argv[0],'/')) == NULL)
    prog = argv[0];
  else
    prog++;

  /* get setup file */
  if (argc < 2) 
    die(stdout, 
	"%s: to few arguments given to program.\n\nNo specification file!\n",
	prog);

  /* since the script is name after a database (or will be) */
  /* the first part of the path is the database user ... */
  if (path_info == NULL) 
    die(stdout, 
	"%s: You must to give an inital path which will be\n\
interpreted as your username when browsing the database.\n\n\
Currenty, you cannot change the username\n\
into something other than given here.", prog);

  if ((f = fopen(argv[1],"r")) == NULL)
    die(stdout,
	"%s: failed to open server specification file:\n\n\t%s\n\n%s\n", 
	prog, argv[1], sys_errlist[errno < sys_nerr ? errno : 0]);
						
  if (fgets(host, sizeof(host), f) == NULL) 
    die(stdout, 
	"%s: failed to read from server specification file:\n\n\t%s\n\n%s\n",
	prog, argv[1], sys_errlist[errno < sys_nerr ? errno : 0]);
  (void) fclose(f);

  /* now, extract host, port and key from it */
  if ((port = strchr(host,':'))    == NULL ||
      (key  = strchr(port+1,':'))    == NULL ||
      (last = strchr(key+1, ':'))    == NULL)
    die(stdout, 
	"%s : something wrong with the format of specification file:\n\n%s\n", 
	prog, argv[1]);

  /* chop of the ':' and make it into string */
  *port++ = *key++ = *last++  = '\0';
  
  /*----------------------------------------------------------------------*/
  if(strcmp(getenv("REQUEST_METHOD"),"GET") == 0) {
    int fd = 0;

    char *p;
    size_t len;
    int    i;
    size_t cnt = 0;

    /* just in case - we are using it in error routine */
    buf[0]='\0';
    if ((fd = tcp_connect(connect_error, host, (int) atol(port))) < 0) 
      die(stdout, buf);

    /* keep only the last part of the script name */
    if (script_name != NULL && (p = strrchr(script_name,'/')) != NULL)
      our_name = p+1;

    len = 10 + strlen(key) + strlen(path_info) + strlen(our_name);

    if (len > sizeof(buf)) {
      if ((p = (char *) malloc(len)) == NULL) {
	(void)close(fd);
	die(stdout, "%s: Failed to allocate memory\n", prog);
      }
    } else 
      p = buf;

    /* We need pass script name back to get path in generated code, */
    /* key is needed for access */
    sprintf(p, "%s/%s%s\n", key, our_name, path_info);
    len = strlen(p);
    if (write(fd, p, len) != len) {
      (void)close(fd);
      die(stdout, 
	  "%s: Failed to write whole path to server\n%s\n",
	  prog, sys_errlist[errno < sys_nerr ? errno : 0]);
    }

    /* just continue coping from socket to stdout ... */
    while((len = read(fd, buf, sizeof(buf))) > 0) 
      for(cnt += len, p=buf; len > 0; len -= i, p+=i)
	if ((i = write(fileno(stdout), p, len)) < 0) {
	  close(fd);
	  die(stdout, 
	      "%s: Failed when writing to stdout\n%s\n",
	      prog, sys_errlist[errno < sys_nerr ? errno : 0]);
	}
      
    close(fd);

    /* nothing was read/written */
    if (cnt == 0)
      die(stdout,
	  "%s: LINCKS database is unable to generate HTML for:\n\n%s%s\n",
	  prog, script_name, path_info);

    exit(0);
  } /* if get */
  /*----------------------------------------------------------------------*/
  else 
    die(stdout, getmsg);

  /* error case */
  exit(1);
}

char x2c(what) 
  char *what;
{
    register char digit;
    digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
    digit *= 16;
    digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
    return(digit);
}

void plustospace(str)
  char *str;
{
    register int x;
    for(x=0;str[x];x++) if(str[x] == '+') str[x] = ' ';
}


